import subprocess
import shlex
import logging
from six import string_types


# FIXME poll_period is not used anywhere
def execute(cmd, shell=False, poll_period=1.0, catch_out=False, executable=None, env=None):
    """Execute UNIX command and wait for its completion

    Args:
        cmd (str or list): command to execute
        shell (bool): invoke inside shell environment
        catch_out (bool): collect process' output
        executable: custom executable for popen

    Returns:
        returncode (int): process return code
        stdout (str): collected process stdout (only if catch_out set to true)
        stderr (str): collected process stderr (only if catch_out set to true)
    """
    # FIXME: move to module level
    log = logging.getLogger(__name__)
    log.info("Starting: %s", cmd)

    stdout = ""
    stderr = ""

    if not shell and isinstance(cmd, string_types):
        cmd = shlex.split(cmd)
    if not executable:
        executable = None

    if catch_out:
        process = subprocess.Popen(
            cmd,
            shell=shell,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            close_fds=True,
            env=env,
            executable=executable)
    else:
        process = subprocess.Popen(cmd, shell=shell, close_fds=True, env=env, executable=executable)

    stdout, stderr = process.communicate()
    if stderr:
        log.error("There were errors:\n%s", stderr)

    if stdout:
        log.debug("Process output:\n%s", stdout)
    returncode = process.returncode
    log.info("Process cmd: %s – exit code: %s", cmd, returncode)
    return returncode, stdout, stderr


# FIXME: remove this dumb popen wrapper
def popen(cmnd):
    return subprocess.Popen(
        cmnd,
        bufsize=0,
        close_fds=True,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE
    )



"""Module provides target monitoring
metrics collector - influxdata's `telegraf` - https://github.com/influxdata/telegraf/
backward compatibility with yandextank's Monitoring module configuration and tools.
"""
import datetime
import fnmatch
import json
import logging
import time
import yaml

from copy import deepcopy

import pkg_resources
from yandextank.plugins.DataUploader.client import LPRequisites
from ...common.interfaces import MonitoringDataListener, AbstractInfoWidget, MonitoringPlugin
from ...common.util import expand_to_seconds, read_resource
from ..Autostop import Plugin as AutostopPlugin, AbstractCriterion
from ..Console import Plugin as ConsolePlugin
from ..Telegraf.collector import MonitoringCollector

from configparser import NoOptionError

logger = logging.getLogger(__name__)


class Plugin(MonitoringPlugin):
    """  resource mon plugin  """

    SECTION = 'telegraf'  # may be redefined to 'monitoring' sometimes.

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.jobno = None
        self.default_target = None
        self.default_config_path = pkg_resources.resource_filename('yandextank.plugins.Telegraf',
                                                                   "config/monitoring_default_config.xml")
        self.process = None
        self.monitoring = MonitoringCollector(
            disguise_hostnames=self.get_option('disguise_hostnames'),
            kill_old=self.get_option('kill_old'))
        self.die_on_fail = True
        self.data_file = None
        self.mon_saver = None
        self._config = None

    @staticmethod
    def get_key():
        return __file__

    def start_test(self):
        if self.monitoring:
            self.monitoring.load_start_time = time.time()
            logger.debug("load_start_time = %s", self.monitoring.load_start_time)

    def get_available_options(self):
        return [
            "config",
            "default_target",
            "ssh_timeout",
            "ssh_key_path",
            "disguise_hostnames"
        ]

    def __detect_configuration(self):
        """
        we need to be flexible in order to determine which plugin's configuration
        specified and make appropriate configs to metrics collector

        :return: SECTION name or None for defaults
        """
        try:
            is_telegraf = self.core.get_option('telegraf', "config")
        except KeyError:
            is_telegraf = None
        try:
            is_monitoring = self.core.get_option('monitoring', "config")
        except KeyError:
            is_monitoring = None

        if is_telegraf and is_monitoring:
            raise ValueError(
                'Both telegraf and monitoring configs specified. '
                'Clean up your config and delete one of them')
        if is_telegraf and not is_monitoring:
            return 'telegraf'
        if not is_telegraf and is_monitoring:
            return 'monitoring'
        if not is_telegraf and not is_monitoring:
            # defaults target logic
            try:
                is_telegraf_dt = self.core.get_option('telegraf')
            except NoOptionError:
                is_telegraf_dt = None
            try:
                is_monitoring_dt = self.core.get_option('monitoring')
            except BaseException:
                is_monitoring_dt = None
            if is_telegraf_dt and is_monitoring_dt:
                raise ValueError(
                    'Both telegraf and monitoring default targets specified. '
                    'Clean up your config and delete one of them')
            if is_telegraf_dt and not is_monitoring_dt:
                return
            if not is_telegraf_dt and is_monitoring_dt:
                self.core.set_option(
                    "telegraf", "default_target", is_monitoring_dt)
            if not is_telegraf_dt and not is_monitoring_dt:
                return

    @property
    def config(self):
        """

        :rtype: str
        """
        if self._config is None:
            value = self.get_option('config')

            if isinstance(value, dict):
                self._config = self._save_config_contents(yaml.dump(value), 'yaml')
            elif value.lower() == "none":
                self.monitoring = None
                self.die_on_fail = False
                self._config = value
            # handle http/https url or file path
            else:
                if value.startswith("<"):
                    config_contents = value
                elif value.lower() == "auto":
                    self.die_on_fail = False
                    config_contents = read_resource(self.default_config_path)
                else:
                    config_contents = read_resource(value)
                self._config = self._save_config_contents(config_contents)

        return self._config

    def _save_config_contents(self, contents, type='xml'):
        outfile = self.core.mkstemp("." + type, "monitoring_")
        self.core.add_artifact_file(outfile)
        with open(outfile, "w") as f:
            f.write(contents)
        return outfile

    def configure(self):
        self.detected_conf = self.__detect_configuration()
        if self.detected_conf:
            logger.info('Detected monitoring configuration: %s', self.detected_conf)
            self.SECTION = self.detected_conf
        self.default_target = self.get_option("default_target", "localhost")
        if self.config.lower() == "none":
            self.monitoring = None
            self.die_on_fail = False
            return

        with open(self.config) as f:
            self.core.add_artifact_to_send(LPRequisites.MONITORING, str(f.read()))

        # FIXME [legacy] backward compatibility with Monitoring module
        # configuration below.
        self.monitoring.ssh_timeout = expand_to_seconds(self.get_option("ssh_timeout", "30s"))
        self.monitoring.ssh_key_path = self.get_option('ssh_key_path', '')
        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
            autostop.add_criterion_class(MetricHigherCriterion)
            autostop.add_criterion_class(MetricLowerCriterion)
        except KeyError:
            logger.debug("No autostop plugin found, not adding metric criterion")

    def prepare_test(self):
        if not self.config or self.config.lower() == 'none':
            return

        if "Phantom" in self.core.job.generator_plugin.__module__:
            phantom = self.core.job.generator_plugin

            info = phantom.get_info()
            if info:
                self.default_target = info.address
                logger.info("Changed monitoring target to %s", self.default_target)

        self.monitoring.config = self.config
        if self.default_target:
            self.monitoring.default_target = self.default_target

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
            console = None
        if console:
            widget = MonitoringWidget(self)
            console.add_info_widget(widget)
            self.monitoring.add_listener(widget)

        try:
            self.monitoring.prepare()
            self.monitoring.start()
            self.add_cleanup(self.monitoring.stop)
            if self.monitoring.agents:
                count = 0
                while not self.monitoring.first_data_received and count < 15 * 5:
                    time.sleep(0.2)
                    self.monitoring.poll()
                    count += 1
        except BaseException:
            logger.exception("Could not start monitoring")
            if self.die_on_fail:
                raise
            else:
                self.monitoring = None

    def add_listener(self, plugin):
        return self.monitoring.add_listener(plugin)

    def is_test_finished(self):
        if self.monitoring:
            monitoring_data = self.monitoring.poll()
            logger.debug("Monitoring got %s lines", len(monitoring_data))
            self.core.publish_monitoring_data(monitoring_data)
        return -1

    def end_test(self, retcode):
        logger.info("Finishing monitoring")
        if self.monitoring:
            self.monitoring.stop()
            for log in self.monitoring.artifact_files:
                self.core.add_artifact_file(log)

            self.core.publish_monitoring_data(self.monitoring.get_rest_data())
        if self.mon_saver:
            self.mon_saver.close()
        return retcode

    def post_process(self, retcode):
        return retcode


class SaveMonToFile(MonitoringDataListener):
    """
    Default listener - saves data to file
    """

    def __init__(self, out_file):
        MonitoringDataListener.__init__(self)
        if out_file:
            self.store = open(out_file, 'w')

    def monitoring_data(self, data):
        self.store.write(json.dumps(data))
        self.store.write('\n')
        self.store.flush()

    def close(self):
        """ close open files """
        logger.debug("Closing monitoring file")
        if self.store:
            self.store.close()


class MonitoringWidget(AbstractInfoWidget, MonitoringDataListener):
    """
    Screen widget
    """

    def __init__(self, owner):
        AbstractInfoWidget.__init__(self)
        self.owner = owner
        self.data = {}
        self.sign = {}
        self.time = {}
        self.max_metric_len = 0

    def get_index(self):
        return 50

    def __handle_data_items(self, host, data):
        """ store metric in data tree and calc offset signs

        sign < 0 is CYAN, means metric value is lower then previous,
        sign > 1 is YELLOW, means metric value is higher then prevoius,
        sign == 0 is WHITE, means initial or equal metric value
        """
        for metric, value in data.items():
            if value == '':
                self.sign[host][metric] = -1
                self.data[host][metric] = value
            else:
                if not self.data[host].get(metric, None):
                    self.sign[host][metric] = 1
                elif float(value) > float(self.data[host][metric]):
                    self.sign[host][metric] = 1
                elif float(value) < float(self.data[host][metric]):
                    self.sign[host][metric] = -1
                else:
                    self.sign[host][metric] = 0
                self.data[host][metric] = "%.2f" % float(value)

    def monitoring_data(self, block):
        # block sample :
        # [{'timestamp': 1480536634,
        #   'data': {
        #     'some.hostname.tld': {
        #       'comment': '',
        #       'metrics': {
        #         'custom:diskio_reads': 0,
        #         'Net_send': 9922,
        #         'CPU_steal': 0,
        #         'Net_recv': 8489
        #       }
        #     }
        #   },
        #   ...
        # }]
        for chunk in block:
            host = next(iter(chunk['data'].keys()))
            self.time[host] = chunk['timestamp']
            # if initial call, we create dicts w/ data and `signs`
            # `signs` used later to paint metrics w/ different colors
            if not self.data.get(host, None):
                self.data[host] = {}
                self.sign[host] = {}
                for key, value in chunk['data'][host]['metrics'].items():
                    self.sign[host][key] = 0
                    self.data[host][key] = value
            else:
                self.__handle_data_items(host, chunk['data'][host]['metrics'])

    def render(self, screen):
        if not self.owner.monitoring:
            return "Monitoring is " + screen.markup.RED + "offline" + screen.markup.RESET
        else:
            res = "Monitoring is " + screen.markup.GREEN + \
                  "online" + screen.markup.RESET + ":\n"
            for hostname, metrics in self.data.items():
                tm_stamp = datetime.datetime.fromtimestamp(
                    float(self.time[hostname])).strftime('%H:%M:%S')
                res += ("   " + screen.markup.CYAN + "%s" + screen.markup.RESET + " at %s:\n") % (hostname, tm_stamp)
                for metric, value in sorted(metrics.items()):
                    if self.sign[hostname][metric] > 0:
                        value = screen.markup.YELLOW + value + screen.markup.RESET
                    elif self.sign[hostname][metric] < 0:
                        value = screen.markup.CYAN + value + screen.markup.RESET
                    res += "      %s%s: %s\n" % (
                        ' ' * (self.max_metric_len - len(metric)),
                        metric.replace('custom:', '').replace('_', ' '), value)

            return res.strip()


class AbstractMetricCriterion(AbstractCriterion, MonitoringDataListener):
    """ Parent class for metric criterion """

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        try:
            self.mon = autostop.core.get_plugin_of_type(Plugin)
            if self.mon.monitoring:
                self.mon.monitoring.add_listener(self)
        except KeyError:
            logger.warning("No monitoring module, mon autostop disabled")
        self.triggered = False
        self.autostop = autostop

        self.host = param_str.split(',')[0].strip()
        self.metric = param_str.split(',')[1].strip()
        self.value_limit = float(param_str.split(',')[2])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[3])
        self.last_second = None
        self.seconds_count = 0

    def monitoring_data(self, _block):
        if self.triggered:
            return

        block = deepcopy(_block)
        for chunk in block:
            host = next(iter(chunk['data'].keys()))
            data = chunk['data'][host]['metrics']

            if not fnmatch.fnmatch(host, self.host):
                continue

            # some magic, converting custom metric names into names that was in
            # config
            for metric_name in tuple(data.keys()):
                if metric_name.startswith('custom:'):
                    config_metric_name = metric_name.replace('custom:', '')
                    data[config_metric_name] = data.pop(metric_name)

            if self.metric not in data or not data[self.metric]:
                data[self.metric] = 0
            logger.debug(
                "Compare %s %s/%s=%s to %s",
                self.get_type_string(), host, self.metric, data[self.metric],
                self.value_limit)
            if self.comparison_fn(float(data[self.metric]), self.value_limit):
                if not self.seconds_count:
                    self.cause_second = self.last_second

                logger.debug(self.explain())

                self.seconds_count += 1
                if self.seconds_count >= self.seconds_limit:
                    logger.debug("Triggering autostop")
                    self.triggered = True
                    return
            else:
                self.seconds_count = 0

    def notify(self, data, stat):
        if self.seconds_count:
            self.autostop.add_counting(self)

        self.last_second = (data, stat)
        return self.triggered

    def comparison_fn(self, arg1, arg2):
        """ comparison function """
        raise NotImplementedError()


class MetricHigherCriterion(AbstractMetricCriterion):
    """ trigger if metric is higher than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return self.RC_METRIC_HIGHER

    @staticmethod
    def get_type_string():
        return 'metric_higher'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is higher than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s > %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 > arg2


class MetricLowerCriterion(AbstractMetricCriterion):
    """ trigger if metric is lower than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return self.RC_METRIC_LOWER

    @staticmethod
    def get_type_string():
        return 'metric_lower'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is lower than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s < %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 < arg2



from ..common.interfaces import AbstractClient

import io
import os
import logging
import json
import threading
try:
    import queue
except ImportError:
    import Queue as queue
import time
from builtins import str

logger = logging.getLogger(__name__)

"""
output artifact sample:

{"dtypes": {"ts": "int64", "value": "float64"}, "type": "metric", "names": ["ts", "value"]}
123	123.123
456	456.456
123	123.123
456	456.456

output meta.json sample:
{
    "metrics": {
        "metric_d12dab4f-e4ef-4c47-89e6-859f73737c64": {
            "dtypes": {
                "ts": "int64",
                "value": "float64"
            },
            "meta": {
                "hostname": "localhost",
                "name": "cpu_usage",
                "some_meta_key": "some_meta_value",
                "type": "metrics"
            },
            "names": [
                "ts",
                "value"
            ],
            "type": "metrics"
        },
    },
    "job_meta": {
        "key": "valueZ",
    }
}
"""


class LocalStorageClient(AbstractClient):
    separator = '\t'
    metrics_meta_fname = 'meta.json'

    def __init__(self, meta, job):
        super(LocalStorageClient, self).__init__(meta, job)
        self.registered_meta = {}

        self.processing_thread = ProcessingThread(self)
        self.processing_thread.daemon = True
        self.processing_thread.start()

    def put(self, data_type, df):
        if df is not None:
            self.pending_queue.put((data_type, df))

    def close(self, test_end):
        self.processing_thread.stop()
        logger.info('Joining local client processing thread...')
        self.processing_thread.join()
        logger.info('Local client finished its work. Artifacts are here %s', self.data_session.artifacts_dir)


class ProcessingThread(threading.Thread):
    """ Process data """
    def __init__(self, client):
        super(ProcessingThread, self).__init__()
        self._finished = threading.Event()
        self._interrupted = threading.Event()
        self.client = client
        self.file_streams = {}

    def __create_artifact(self, metric_full_name):
        self.file_streams[metric_full_name] = io.open(
            os.path.join(
                self.client.data_session.artifacts_dir, "{}.data".format(metric_full_name)
            ),
            mode='w'
        )

    def run(self):
        while not self._interrupted.is_set():
            self.__process_pending_queue()
        logger.info(
            'File writer thread interrupted, finishing work and trying to write the rest of data, qsize: %s',
            self.client.pending_queue.qsize()
        )
        self.__process_pending_queue()
        self.__close_files_and_dump_meta()
        self._finished.set()

    def __process_pending_queue(self):
        try:
            data_type, incoming_df = self.client.pending_queue.get_nowait()
            df = incoming_df.copy()[data_type.columns]
        except queue.Empty:
            time.sleep(1)
        else:
            for metric_local_id, df_grouped_by_id in df.groupby(level=0, sort=False):
                metric = self.client.data_session.manager.get_metric_by_id(metric_local_id)
                if not metric:
                    logger.warning('Received unknown metric id: %s', metric_local_id)
                    return
                metric_full_name = '{}_{}'.format(data_type.table_name, metric_local_id)
                if metric_full_name not in self.file_streams:
                    logger.debug('Creating artifact file for %s', metric_full_name)
                    self.__create_artifact(metric_full_name)
                    this_metric_meta = {
                        'type': data_type.__name__,
                        'names': data_type.columns,
                        'dtypes': df.dtypes.apply(lambda x: x.name).to_dict(),
                        'meta': metric.meta
                    }
                    self.client.registered_meta[metric_full_name] = this_metric_meta
                    artifact_file_header = json.dumps(this_metric_meta)
                    self.file_streams[metric_full_name].write("%s\n" % artifact_file_header)
                csv_data = df_grouped_by_id.to_csv(
                    sep=self.client.separator,
                    header=False,
                    index=False,
                    na_rep="",
                    columns=data_type.columns,
                )
                try:
                    self.file_streams[metric_full_name].write(
                        str(csv_data)
                    )
                    self.file_streams[metric_full_name].flush()
                except ValueError:
                    logger.warning('Failed to write metrics to file, maybe file is already closed?', exc_info=True)

    def is_finished(self):
        return self._finished

    def __close_files_and_dump_meta(self):
        [self.file_streams[file_].close() for file_ in self.file_streams]
        with open(os.path.join(self.client.data_session.artifacts_dir, self.client.metrics_meta_fname), 'w') as meta_f:
            json.dump(
                {"metrics": self.client.registered_meta, "job_meta": self.client.meta},
                meta_f,
                indent=4,
                sort_keys=True
            )

    def stop(self):
        self._interrupted.set()



import itertools
import time
from ...common import util


def uts(dt):
    return int(time.mktime(dt.timetuple()))


class Decoder(object):
    """
    Decode metrics incoming from tank into points for OpenTSDB

    Parameters
    ----------
    parent_tags : dict
        common per-test tags
    tank_tag : str
        tank identifier tag
    uuid : str
        test id tag
    labeled : bool
        detailed stats for each label
    histograms : bool
        response time histograms measurements
    """
    def __init__(self, tank_tag, uuid, parent_tags, labeled, histograms):
        self.labeled = labeled
        initial_tags = {"tank": tank_tag, "uuid": uuid}
        initial_tags.update(parent_tags)
        self.tags = initial_tags
        self.histograms = histograms

    def set_uuid(self, id_):
        self.tags['uuid'] = id_

    def decode_monitoring(self, data):
        """
        The reason why we have two separate methods for monitoring
        and aggregates is a strong difference in incoming data.
        """
        points = list()
        for second_data in data:
            for host, host_data in second_data["data"].items():
                points.append(
                    self.__make_points(
                        "monitoring",
                        {
                            "host": host,
                            "comment": host_data.get("comment")
                        },
                        second_data["timestamp"],
                        {
                            # cast int to float. avoid
                            # https://github.com/yandex/yandex-tank/issues/776
                            metric:
                            float(value) if isinstance(value, int) else value
                            for metric, value in
                            host_data["metrics"].items()
                        }))
        return list(itertools.chain(*points))

    def decode_aggregates(self, aggregated_data, gun_stats, prefix):
        ts = aggregated_data["ts"]
        points = list()
        # stats overall w/ __OVERALL__ label
        points += self.__make_points_for_label(
            ts, aggregated_data["overall"], "__OVERALL__", prefix, gun_stats)
        # detailed stats per tag
        if self.labeled:
            for label, aggregated_data_by_tag in aggregated_data[
                    "tagged"].items():
                points += self.__make_points_for_label(
                    ts, aggregated_data_by_tag, label, prefix, gun_stats)
        return points

    def __make_points_for_label(self, ts, data, label, prefix, gun_stats):
        """x
        Make a set of points for `this` label

        overall_quantiles, overall_meta, net_codes, proto_codes, histograms
        """
        label_points = list()

        label_points.extend((
            # overall quantiles for label
            self.__make_points(
                prefix + "overall_quantiles", {"label": label}, ts,
                self.__make_quantile_fields(data)),
            # overall meta (gun status) for label
            self.__make_points(
                prefix + "overall_meta", {"label": label}, ts,
                self.__make_overall_meta_fields(data, gun_stats)),
            # net codes for label
            self.__make_points(
                prefix + "net_codes", {"label": label},
                ts,
                self.__make_netcodes_fields(data),
                field_lookup_table=util.NET),
            # proto codes for label
            self.__make_points(
                prefix + "proto_codes", {"label": label},
                ts,
                self.__make_protocodes_fields(data),
                field_lookup_table=util.HTTP)))
        # histograms, one row for each bin
        if self.histograms:
            for bin_, count in zip(data["interval_real"]["hist"]["bins"],
                                   data["interval_real"]["hist"]["data"]):
                label_points.append(
                    self.__make_points(
                        prefix + "histograms", {"label": label}, ts, {
                            "bin": bin_,
                            "count": count
                        }))
        return list(itertools.chain(*label_points))

    @staticmethod
    def __make_quantile_fields(data):
        return {
            'q' + str(q): value / 1000.0
            for q, value in zip(
                data["interval_real"]["q"]["q"], data["interval_real"]["q"]
                ["value"])
        }

    @staticmethod
    def __make_overall_meta_fields(data, stats):
        return {
            "active_threads":
            stats["metrics"]["instances"],
            "RPS":
            data["interval_real"]["len"],
            "planned_requests":
            float(stats["metrics"]["reqps"]),
            "avg_rt":
            float(data['interval_real']['total']) / data['interval_real']['len'] / 1000.0,
            "min":
            data['interval_real']['min'] / 1000.0,
            "max":
            data['interval_real']['max'] / 1000.0
        }

    @staticmethod
    def __make_netcodes_fields(data):
        return {
            int(code): int(cnt)
            for code, cnt in data["net_code"]["count"].items()
        }

    @staticmethod
    def __make_protocodes_fields(data):
        return {
            int(code): int(cnt)
            for code, cnt in data["proto_code"]["count"].items()
        }

    def __make_points(
            self, metric, additional_tags, ts, fields, field_lookup_table={}):
        """
        Parameters
        ----------
        metric : string
            metric type (e.g. monitoring, overall_meta, net_codes, proto_codes, overall_quantiles)
        additional_tags : dict
            custom additional tags for this points
        ts : integer
            timestamp
        fields : dict
            opentsdb fields tag

        Returns
        -------
        dict
            points for OpenTSDB client
        """
        result = []

        for key, value in fields.items():
            tags = self.tags.copy()
            tags.update(additional_tags)
            tags["field"] = str(key)
            if field_lookup_table.get(key):
                tags["field_label"] = field_lookup_table.get(key).replace(" ", "_")
            result.append({
                "metric": metric,
                "tags": tags,
                "timestamp": int(ts),
                "value": value,
            })

        return result



import copy
import json
import logging
import re
import time

from yandextank.common.util import expand_to_seconds, expand_to_milliseconds
from ...common.interfaces import AbstractCriterion
from ..Phantom import Plugin as PhantomPlugin
from ..Pandora import Plugin as PandoraPlugin

logger = logging.getLogger(__name__)


class AvgTimeCriterion(AbstractCriterion):
    """ average response time criterion """

    @staticmethod
    def get_type_string():
        return 'time'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.rt_limit = expand_to_milliseconds(params[0])
        self.seconds_limit = expand_to_seconds(params[1])
        self.autostop = autostop
        self.tag = params[2].strip() if len(params) == 3 else None

    def notify(self, data, stat):
        rt_total, requests_number = self.parse_data(data)
        rt_actual = rt_total / 1000.0 / requests_number

        if rt_actual > self.rt_limit:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def parse_data(self, data):
        requests_number = data["overall"]["interval_real"]["len"]
        # Parse data for specific tag if it is present
        if self.tag:
            if data["tagged"].get(self.tag):
                rt_total = data["tagged"][self.tag]["interval_real"]["total"]
                requests_number = data["tagged"][self.tag]["interval_real"]["len"]
            # rt_total=0 if current tag differs from selected one
            else:
                rt_total = 0
        # Parse data for overall
        else:
            rt_total = data["overall"]["interval_real"]["total"]
        return rt_total, requests_number

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "Average response time higher than %(limit)sms for %(seconds_count)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'limit': self.rt_limit,
            'seconds_count': self.seconds_count,
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "Avg Time >%(limit)sms for %(seconds_count)s/%(seconds_limit)ss" % items, \
            float(self.seconds_count) / self.seconds_limit


class HTTPCodesCriterion(AbstractCriterion):
    """ HTTP codes criterion """

    @staticmethod
    def get_type_string():
        return 'http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses:
                matched_responses = float(matched_responses) / total_responses
            else:
                matched_responses = 0

        logger.debug("HTTP codes matching mask %s: %s/%s", self.codes_mask, matched_responses, self.level)

        if matched_responses >= self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def parse_data(self, data):
        # Parse data for specific tag
        if self.tag:
            if data["tagged"].get(self.tag):
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
                matched_responses = self.count_matched_codes(
                    self.codes_regex, data["tagged"][self.tag]["proto_code"]["count"])
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Parse data for overall
        else:
            matched_responses = self.count_matched_codes(self.codes_regex, data["overall"]["proto_code"]["count"])
            total_responses = data["overall"]["interval_real"]["len"]
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_HTTP

    def get_level_str(self):
        """ format level str """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "%(code)s codes count higher than %(level)s for %(seconds_count)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_count': self.seconds_count,
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "HTTP %(code)s>%(level)s for %(seconds_count)s/%(seconds_limit)ss" % items, \
            float(self.seconds_count) / self.seconds_limit


class NetCodesCriterion(AbstractCriterion):
    """ Net codes criterion """

    @staticmethod
    def get_type_string():
        return 'net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses:
                matched_responses = float(matched_responses) / total_responses
            else:
                matched_responses = 0

        logger.debug("Net codes matching mask %s: %s/%s", self.codes_mask, matched_responses, self.level)

        if matched_responses >= self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def parse_data(self, data):
        # Count data for specific tag if it's present
        if self.tag:
            if data["tagged"].get(self.tag):
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
                code_count = data["tagged"][self.tag]["net_code"]["count"]
                codes = copy.deepcopy(code_count)
                if '0' in codes:
                    codes.pop('0')
                matched_responses = self.count_matched_codes(self.codes_regex, codes)
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Count data for overall
        else:
            code_count = data["overall"]["net_code"]["count"]
            total_responses = data["overall"]["interval_real"]["len"]
            codes = copy.deepcopy(code_count)
            if '0' in codes:
                codes.pop('0')
            matched_responses = self.count_matched_codes(self.codes_regex, codes)
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_NET

    def get_level_str(self):
        """ format level str """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "%(code)s net codes count higher than %(level)s for %(seconds_count)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_count': self.seconds_count,
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "Net %(code)s>%(level)s for %(seconds_count)s/%(seconds_limit)ss" % items, \
            float(self.seconds_count) / self.seconds_limit


class QuantileCriterion(AbstractCriterion):
    """ quantile criterion """

    @staticmethod
    def get_type_string():
        return 'quantile'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.quantile = float(params[0])
        self.rt_limit = expand_to_milliseconds(params[1])
        self.seconds_limit = expand_to_seconds(params[2])
        self.autostop = autostop
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        quantiles = self.parse_data(data)
        logger.debug('Autostop quantiles for ts %s: %s', data['ts'], quantiles)
        if self.quantile not in quantiles.keys():
            logger.warning("No quantile %s in %s", self.quantile, quantiles)
        if self.quantile in quantiles.keys() and quantiles[self.quantile] / 1000.0 > self.rt_limit:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            self.seconds_count += 1
            logger.debug(self.explain())
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def parse_data(self, data):
        # Parse data for specific tag
        if self.tag:
            if data["tagged"].get(self.tag):
                quantile_values = data["tagged"][self.tag]["interval_real"]["q"]["value"]
            # quantile_values empty if current tag differs from selected one
            else:
                quantile_values = []
        # Parse data for overall
        else:
            quantile_values = data["overall"]["interval_real"]["q"]["value"]
        quantiles = dict(zip(data["overall"]["interval_real"]["q"]["q"], quantile_values))
        return quantiles

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "Percentile %(percentile)s higher than %(limit)sms for %(seconds_count)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'percentile': self.quantile,
            'limit': self.rt_limit,
            'seconds_count': self.seconds_count,
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "%(percentile)s%% >%(limit)sms for %(seconds_count)s/%(seconds_limit)ss" % items, \
            float(self.seconds_count) / self.seconds_limit


class SteadyCumulativeQuantilesCriterion(AbstractCriterion):
    """ quantile criterion """

    @staticmethod
    def get_type_string():
        return 'steady_cumulative'

    def __init__(self, autostop, param_str):
        raise NotImplementedError
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.quantile_hash = ""
        self.seconds_limit = expand_to_seconds(param_str.split(',')[0])
        self.autostop = autostop

    def notify(self, data, stat):
        quantiles = dict(
            zip(data["overall"]["q"]["q"], data["overall"]["q"]["values"]))
        quantile_hash = json.dumps(quantiles)
        logging.debug("Cumulative quantiles hash: %s", quantile_hash)
        if self.quantile_hash == quantile_hash:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        self.quantile_hash = quantile_hash
        return False

    def get_rc(self):
        return self.RC_STEADY

    def explain(self):
        return "Cumulative percentiles are steady for %ss" % self.seconds_count

    def widget_explain(self):
        items = (self.seconds_count, self.seconds_limit)
        return "Steady for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class TimeLimitCriterion(AbstractCriterion):
    """ time limit criterion """

    @staticmethod
    def get_type_string():
        return 'limit'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.start_time = time.time()
        self.end_time = time.time()
        self.time_limit = expand_to_seconds(param_str)

    def notify(self, data, stat):
        self.end_time = time.time()
        return (self.end_time - self.start_time) > self.time_limit

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        return "Test time elapsed. Limit: %(limit)ss" % self.get_criterion_parameters()

    def get_criterion_parameters(self):
        parameters = {
            'limit': self.time_limit,
            'actual': self.end_time - self.start_time
        }
        return parameters

    def widget_explain(self):
        return "Time limit: %(limit)ss, actual time: %(actual)ss" % self.get_criterion_parameters()


class UsedInstancesCriterion(AbstractCriterion):
    """
    Autostop criterion, based on active instances count
    """

    @staticmethod
    def get_type_string():
        return 'instances'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.autostop = autostop
        self.threads_limit = 0

        level_str = param_str.split(',')[0].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[1])

        if self.autostop.core.get_option('phantom', 'enabled', False):
            phantom = autostop.core.get_plugin_of_type(PhantomPlugin)
            info = phantom.get_info()
            if info:
                self.threads_limit = info.instances
            if not self.threads_limit:
                raise ValueError(
                    "Cannot create 'instances' criterion"
                    " with zero instances limit")
        elif self.autostop.core.get_option('pandora', 'enabled', False):
            config = autostop.core.get_plugin_of_type(PandoraPlugin).config_contents
            for pool in config["pools"]:
                if pool["startup"]["type"] == "once":
                    self.threads_limit += pool["startup"]["times"]
                else:
                    raise ValueError(
                        "Cannot create 'instances' criterion"
                        " startup type must be 'once'")
        else:
            raise ValueError(
                "Cannot create relative 'instances' criterion"
                " either phantom or pandora must be enabled")

    def notify(self, data, stat):
        threads = stat["metrics"]["instances"]
        if self.is_relative:
            threads = float(threads) / self.threads_limit
        if threads > self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_INSTANCE

    def get_level_str(self):
        """
        String value for instances level
        """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = (self.get_level_str(), self.seconds_count)
        return "Testing threads (instances) utilization higher than %s for %ss" % items

    def widget_explain(self):
        items = (self.get_level_str(), self.seconds_count, self.seconds_limit)
        return "Instances >%s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit



import importlib
import os
import re
import sys
import uuid

import logging

import glob
import pkg_resources
import yaml
from cerberus.validator import Validator
from functools import reduce

from yandextank.common.util import read_resource, recursive_dict_update
logger = logging.getLogger(__name__)


class ValidationError(Exception):
    MSG_TEMPLATE = """Validation error:\n{}"""

    def __init__(self, errors):
        self.errors = errors
        self.message = self.MSG_TEMPLATE.format(yaml.dump(self.errors))

    def __str__(self):
        return self.message


def load_yaml_schema(path):
    # DEFAULT_FILENAME = 'schema.yaml'
    file_content = read_resource(path)
    return yaml.safe_load(file_content)


def load_py_schema(package):
    schema_module = importlib.import_module(package + '.config.schema')
    return schema_module.SCHEMA


def load_plugin_schema(package, ignore_import_error=False):
    try:
        return load_yaml_schema(
            pkg_resources.resource_filename(
                package, 'config/schema.yaml'))
    except IOError:
        try:
            return load_py_schema(package)
        except ImportError:
            logger.error(
                "Could not find schema for %s (should be located in config/ directory of a plugin)",
                package)
            raise IOError('No schema found for plugin %s' % package)
    except ImportError:
        if 'aggregator' in package.lower():
            logger.exception('Plugin Aggregator is now deprecated, please remove this section from your config.')
        if ignore_import_error:
            return None
        raise ValidationError({'package': ['No module named {}'.format(package)]})


def load_schema(directory, filename=None):
    try:
        return load_yaml_schema(directory)
    except IOError:
        try:
            return load_py_schema(directory)
        except ImportError:
            raise IOError(
                'Neither .yaml nor .py schema found in %s' %
                directory)


class PatchedValidator(Validator):

    def _validate_description(self, description, field, value):
        """ {'type': 'string'} """
        pass

    def _validate_values_description(self, values_description, field, value):
        """ {'type': 'dict'} """
        pass

    def _validate_tutorial_link(self, tutorial_link, field, value):
        """ {'type': 'string'} """
        pass

    def _validate_examples(self, examples, field, value):
        """ {'type': 'dict'} """
        pass

    @staticmethod
    def is_number(value):
        try:
            float(value)
            return True
        except ValueError:
            return False

    def validate_duration(self, field, duration):
        '''
        2h
        2h5m
        5m
        180
        1h4m3
        :param duration:
        :return:
        '''
        DURATION_RE = r'^(\d+d)?(\d+h)?(\d+m)?(\d+s?)?$'
        if not re.match(DURATION_RE, duration):
            self._error(field, 'Load duration examples: 2h30m; 5m15; 180')

    def _check_with_load_scheme(self, field, value):
        '''
        step(10,200,5,180)
        step(5,50,2.5,5m)
        line(22,154,2h5m)
        step(5,50,2.5,5m) line(22,154,2h5m)
        const(10,1h4m3s)
        :param field:
        :param value:
        :return:
        '''
        # stpd file can be any value
        if self.document['load_type'] in 'stpd_file':
            return

        PRIMARY_RE = r'(step|line|const)\((.+?)\)'
        N_OF_ARGS = {
            'step': 4,
            'line': 3,
            'const': 2,
        }
        matches = re.findall(PRIMARY_RE, value)
        if len(matches) == 0:
            self._error(field, 'Should match one of the following patterns: step(...) / line(...) / const(...)')
        else:
            for match in matches:
                curve, params_str = match
                params = [v.strip() for v in params_str.split(',')]
                # check number of arguments
                if not len(params) == N_OF_ARGS[curve]:
                    self._error(field, '{} load scheme: expected {} arguments, found {}'.format(curve,
                                                                                                N_OF_ARGS[curve],
                                                                                                len(params)))
                # check arguments' types
                for param in params[:-1]:
                    if not self.is_number(param):
                        self._error(field, 'Argument {} in load scheme should be a number'.format(param))
                self.validate_duration(field, params[-1])


class TankConfig(object):
    DYNAMIC_OPTIONS = {
        'uuid': lambda: str(uuid.uuid4()),
        'pid': lambda: os.getpid(),
        'cmdline': lambda: ' '.join(sys.argv)
    }

    def __init__(
            self,
            configs,
            with_dynamic_options=True,
            core_section='core',
            error_output=None,
            skip_base_cfgs=True,
            plugins_implicit_enabling=False,
            skip_unknown_plugins=False,
    ):
        """
        :param configs: list of configs dicts
        :param with_dynamic_options: insert uuid, pid, and other DYNAMIC_OPTIONS
        :param core_section: name of core section in config
        :param error_output: file to output error messages
        :param skip_base_cfgs: not merge input configs with base configs
        :param plugins_implicit_enabling: enable plugin if it`s declared in config without "enabled: false" explicitly
        """
        if not isinstance(configs, list):
            configs = [configs]
        self.raw_config_dict = self.__construct_config(configs, skip_base_cfgs, plugins_implicit_enabling)
        if self.raw_config_dict.get(core_section) is None:
            self.raw_config_dict[core_section] = {}
        self.with_dynamic_options = with_dynamic_options
        self.skip_unknown_plugins = skip_unknown_plugins
        self.CORE_SECTION = core_section
        self._validated = None
        self._plugins = None
        self.ERROR_OUTPUT = error_output
        self.BASE_SCHEMA = load_yaml_schema(pkg_resources.resource_filename('yandextank.core', 'config/schema.yaml'))
        self.PLUGINS_SCHEMA = load_yaml_schema(pkg_resources.resource_filename('yandextank.core', 'config/plugins_schema.yaml'))

    def get_configinitial(self):
        return self.raw_config_dict

    def validate(self):
        if not self._validated:
            self._validated = ValidatedConfig(self.__validate(), self.BASE_SCHEMA)
        return self._validated, self.raw_config_dict

    @property
    def validated(self):
        if not self._validated:
            try:
                self._validated = self.__validate()
            except ValidationError as e:
                if self.ERROR_OUTPUT:
                    with open(self.ERROR_OUTPUT, 'w') as f:
                        yaml.dump(e.errors, f)
                raise
        return self._validated

    def save_raw(self, filename):
        with open(filename, 'w') as f:
            yaml.dump(self.raw_config_dict, f)

    def __construct_config(self, user_configs, skip_base_cfgs, plugins_implicit_enabling):
        user_config = self.__merge_configs(user_configs)

        core_base_config = load_core_base_cfg()
        local_base_configs = load_local_base_cfgs()
        base_config = self.__merge_configs([core_base_config] + local_base_configs)

        # YANDEXTANK-764
        if plugins_implicit_enabling:
            user_config = self.__enable_plugins_implicitly(user_config, plugins=base_config)
        if skip_base_cfgs:
            return user_config

        # YANDEXTANK-715: for backward compatibility purposes
        user_config = self.__patch_phantom_implicit_enabled_true(user_config)
        config = self.__merge_configs([base_config, user_config])
        return config

    @staticmethod
    def __merge_configs(configs):
        logger.debug('Configs: {}'.format(configs))
        configs = list(filter(None, configs))
        return reduce(recursive_dict_update, configs, {})

    @staticmethod
    def __enable_plugins_implicitly(config, plugins):
        for name, content in config.items():
            if name not in plugins:
                continue
            content = content or {}
            content['enabled'] = content.get('enabled', True)
            default_content = plugins.get(name) or {}
            if plugin_package := default_content.get('package'):
                content['package'] = content.get('package', plugin_package)
            config[name] = content
        return config

    @staticmethod
    def __patch_phantom_implicit_enabled_true(config):
        if 'phantom' in config:
            content = config.get('phantom') or {}
            content['enabled'] = content.get('enabled', True)
            config['phantom'] = content
        return config

    def __parse_enabled_plugins(self):
        """
        :returns: [(plugin_name, plugin_package, plugin_config), ...]
        :rtype: list of tuple
        """
        return [
            (
                plugin_name,
                plugin['package'],
                plugin) for plugin_name,
            plugin in self.raw_config_dict.items() if (
                plugin_name not in self.BASE_SCHEMA) and isinstance(
                plugin,
                dict) and plugin.get('enabled')]

    def __validate(self):
        core_validated = self.__validate_core()
        # plugins:
        errors = {}
        results = {}
        for plugin_name, package, config in self.__parse_enabled_plugins():
            try:
                schema = load_plugin_schema(package, self.skip_unknown_plugins)
                results[plugin_name] = self.__validate_unknown(config) if schema is None else self.__validate_plugin(config, schema)
            except ValidationError as e:
                errors[plugin_name] = e.errors
        if len(errors) > 0:
            raise ValidationError(dict(errors))

        for plugin_name, plugin_conf in results.items():
            core_validated[plugin_name] = plugin_conf
        return core_validated

    def __validate_core(self):
        v = PatchedValidator(allow_unknown=self.PLUGINS_SCHEMA)
        result = v.validate(self.raw_config_dict, self.BASE_SCHEMA)
        if not result:
            errors = v.errors
            for key, value in tuple(errors.items()):
                if 'must be of dict type' in value:
                    errors[key] = ['unknown field']
            raise ValidationError(errors)
        normalized = v.normalized(self.raw_config_dict)
        return self.__set_core_dynamic_options(
            normalized) if self.with_dynamic_options else normalized

    def __validate_unknown(self, config):
        v = PatchedValidator(self.PLUGINS_SCHEMA['schema'], allow_unknown=True)
        # .validate() makes .errors as side effect if there's any
        if not v.validate(config):
            raise ValidationError(v.errors)
        return config

    def __validate_plugin(self, config, schema):
        schema.update(self.PLUGINS_SCHEMA['schema'])
        v = PatchedValidator(schema, allow_unknown=False)
        # .validate() makes .errors as side effect if there's any
        if not v.validate(config):
            raise ValidationError(v.errors)
        # .normalized() returns config with defaults
        return v.normalized(config)

    def __set_core_dynamic_options(self, config):
        for option, setter in self.DYNAMIC_OPTIONS.items():
            try:
                config[self.CORE_SECTION][option] = setter()
            except KeyError:
                config[self.CORE_SECTION] = {option: setter()}
        return config

    def __str__(self):
        return yaml.dump(self.raw_config_dict)


class ValidatedConfig(object):
    def __init__(self, validated, base_schema):
        """

        :type validated: dict
        """
        self.validated = validated
        self.base_schema = base_schema
        self._plugins = None

    @property
    def plugins(self):
        """
            :returns: [(plugin_name, plugin_package, plugin_config), ...]
            :rtype: list of tuple
        """
        if not self._plugins:
            self._plugins = [
                (plugin_name,
                 plugin_cfg['package'],
                 plugin_cfg) for plugin_name, plugin_cfg in self.validated.items() if (
                    plugin_name not in self.base_schema) and plugin_cfg['enabled']]
        return self._plugins

    def get_option(self, section, option, default=None):
        try:
            return self.validated[section][option]
        except KeyError:
            if default is not None:
                return default
            raise

    def __bool__(self):
        return len(self.validated) > 0

    def dump(self, path):
        with open(path, 'w') as f:
            yaml.dump(self.validated, f)

    def __str__(self):
        return yaml.dump(self.validated)


def load_core_base_cfg():
    return load_yaml_cfg(pkg_resources.resource_filename('yandextank.core', 'config/00-base.yaml'))


def load_local_base_cfgs():
    return cfg_folder_loader('/etc/yandex-tank')


def cfg_folder_loader(path):
    """
    :type path: str
    """
    CFG_WILDCARD = '*.yaml'
    return [load_yaml_cfg(filename) for filename in sorted(glob.glob(os.path.join(path, CFG_WILDCARD)))]


def load_yaml_cfg(cfg_filename):
    """
    :type cfg_filename: str
    """
    cfg_yaml = yaml.load(read_resource(cfg_filename), Loader=yaml.FullLoader)
    if not isinstance(cfg_yaml, dict):
        raise ValidationError('Wrong config format, should be a yaml')
    return cfg_yaml



'''
Utilities: parsers, converters, etc.
'''
import re
import logging
import math
from itertools import islice

from .module_exceptions import StepperConfigurationError

logging.getLogger("requests").setLevel(logging.WARNING)


def take(number, iter):
    return list(islice(iter, 0, number))


def parse_duration(duration):
    '''
    Parse duration string, such as '3h2m3s' into milliseconds

    >>> parse_duration('3h2m3s')
    10923000

    >>> parse_duration('0.3s')
    300

    >>> parse_duration('5')
    5000
    '''
    _re_token = re.compile("([0-9.]+)([dhms]?)")

    def parse_token(time, multiplier):
        multipliers = {
            'd': 86400,
            'h': 3600,
            'm': 60,
            's': 1,
        }
        if multiplier:
            if multiplier in multipliers:
                return int(float(time) * multipliers[multiplier] * 1000)
            else:
                raise StepperConfigurationError('Failed to parse duration: %s' % duration)
        else:
            return int(float(time) * 1000)

    return sum(parse_token(*token) for token in _re_token.findall(duration))


def solve_quadratic(a, b, c):
    '''
    >>> solve_quadratic(1.0, 2.0, 1.0)
    (-1.0, -1.0)
    '''
    discRoot = math.sqrt((b * b) - 4 * a * c)
    root1 = (-b - discRoot) / (2 * a)
    root2 = (-b + discRoot) / (2 * a)
    return (root1, root2)


def s_to_ms(f_sec):
    return int(f_sec * 1000.0)


def proper_round(n):
    """
    rounds float to closest int
    :rtype: int
    :param n: float
    """
    return int(n) + (n / abs(n)) * int(abs(n - int(n)) >= 0.5) if n != 0 else 0



import collections
import time
import logging
import os
import re
import yaml

# TODO: rename to format_http_request
from threading import Lock


def pretty_print(req):
    return '{header}\n{query}\n{http_headers}\n\n{body}\n{footer}'.format(
        header='-----------QUERY START-----------',
        query=req.method + ' ' + req.url,
        http_headers='\n'.join('{}: {}'.format(k, v) for k, v in req.headers.items()),
        body=req.body,
        footer='-----------QUERY END-----------'
    )


def recursive_dict_update(d1, d2):
    for k, v in d2.items():
        if isinstance(v, collections.Mapping):
            r = recursive_dict_update(d1.get(k, {}), v)
            d1[k] = r
        else:
            d1[k] = d2[k]
    return d1


def log_time_decorator(func):
    """
    logs func execution time
    :param func:
    :return:
    """
    def timed(*args, **kwargs):
        start = time.time()
        res = func(*args, **kwargs)
        logging.debug('TIMER {}: {}'.format(func.__name__, round(time.time() - start, 3)))
        return res

    return timed


class thread_safe_property(object):
    # credits to https://stackoverflow.com/a/39217007/3316574
    def __init__(self, func):
        self._func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self._lock = Lock()

    def __get__(self, obj, klass=None):
        if obj is None:
            return None
        # __get__ may be called concurrently
        with self._lock:
            # another thread may have computed property value
            # while this thread was in __get__
            if self.__name__ not in obj.__dict__:
                # none computed `_func` yet, do so (under lock) and set attribute
                obj.__dict__[self.__name__] = self._func(obj)
        # by now, attribute is guaranteed to be set,
        # either by this thread or another
        return obj.__dict__[self.__name__]


def expandvars(path, default=None):
    if default is None:
        return os.path.expandvars(path)

    # matches expressions like ${VALUE} where VALUE is parsed to group 1
    reVar = r'\$\{([^}]*)\}'

    def replace_var(m: re.Match):
        return os.environ.get(m.group(1), default)

    return re.sub(reVar, replace_var, path)


def env_constructor(loader, node):
    return expandvars(node.value, default='')


class YamlEnvSubstConfigLoader(yaml.SafeLoader):
    pass


env_matcher = re.compile(r'.*\$\{([^}^{]+)\}.*')
YamlEnvSubstConfigLoader.add_implicit_resolver('!env', env_matcher, None)
YamlEnvSubstConfigLoader.add_constructor('!env', constructor=env_constructor)



import threading
import time
import six
import pandas as pd
import logging
from yandextank.contrib.netort.netort.data_manager.common.interfaces import Aggregated


if six.PY3:
    from queue import Empty
else:  # six.PY2
    from Queue import Empty


logger = logging.getLogger(__name__)


class MetricsRouter(threading.Thread):
    """
    Drain incoming queue, concatenate dataframes by metric type and process to callbacks
    callback receives resulting dataframe
    """

    # TODO: MetricsRouter should not know anything about DataManager. Pass source and subscribers directly.
    def __init__(self, manager, aggregator_buffer_size=10):
        """
        :param aggregator_buffer_size: seconds
        :type aggregator_buffer_size: int
        :type manager: yandextank.contrib.netort.netort.data_manager.DataManager
        """
        super(MetricsRouter, self).__init__()
        self.aggregator_buffer_size = aggregator_buffer_size
        self.manager = manager
        self._finished = threading.Event()
        self._stopped = threading.Event()
        self._interrupted = threading.Event()
        self.daemon = True  # just in case, bdk+ytank stuck w/o this at join of Drain thread
        self.__buffer = {}

    def run(self):
        while not self._stopped.is_set():
            self.__route()
        logger.info('Router received interrupt signal, routing rest of the data. Qsize: %s',
                    self.manager.routing_queue.qsize())
        while self.manager.routing_queue.qsize() > 1 and not self._interrupted.is_set():
            self.__route()
        self.__route(last_piece=True)
        logger.info('Router finished its work')
        self._finished.set()

    def _from_buffer(self, metric_data, last_piece):
        """
        :type metric_data: yandextank.contrib.netort.netort.data_manager.common.interfaces.MetricData
        :rtype: pd.DataFrame
        """
        buffered = self.__buffer.pop(metric_data.local_id, None)
        df = pd.concat([buffered, metric_data.df]) if buffered is not None else metric_data.df
        if not last_piece:
            cut, new_buf = df[df.second < df.second.max() - Aggregated.buffer_size], \
                df[df.second >= df.second.max() - Aggregated.buffer_size]
            self.__buffer[metric_data.local_id] = new_buf
            return cut
        else:
            return df

    @staticmethod
    def __process_df(df, dtype, local_id):
        processed_df = dtype.processor(df)
        processed_df.loc[:, 'metric_local_id'] = local_id
        return processed_df

    def __route(self, last_piece=False):
        try:
            # metric_data contains only one metric and one/several data_types
            metric_data = self.manager.routing_queue.get_nowait()
        except Empty:
            return
        routed_data = {}
        from_buffer = self._from_buffer(metric_data, last_piece) if metric_data.is_aggregated else metric_data.df
        for dtype in metric_data.data_types:
            unprocessed = from_buffer if dtype.is_aggregated() else metric_data.df
            if not unprocessed.empty:
                processed = self.__process_df(unprocessed, dtype, metric_data.local_id)
                if not processed.empty:
                    routed_data.setdefault(dtype, []).append(processed)

        if last_piece:
            for metric_local_id, df in self.__buffer.items():
                d_types = self.manager.metrics[metric_local_id].data_types
                for dtype in [dt for dt in d_types if dt.is_aggregated()]:
                    processed = self.__process_df(df, dtype, metric_local_id)
                    routed_data.setdefault(dtype, []).append(processed)

        for dtype, dfs in routed_data.items():
            for df in dfs:
                for subscriber in self.manager.callbacks:
                    callback = self.manager.subscribers[subscriber]
                    callback(dtype, df)
            routed_data[dtype] = pd.concat(dfs, sort=False)

        if not self.manager.callbacks:
            logger.debug('No subscribers/callbacks for metrics yet... skipped metrics')
            time.sleep(1)
            return

    def wait(self, timeout=None):
        self._finished.wait(timeout=timeout)

    def close(self):
        self._stopped.set()

    def interrupt(self):
        self.close()
        logger.debug('Routing interrupted')
        self._interrupted.set()



"""
Logging initialization AND signals (but why?!)

TODO:
    * move signals to another module
"""
import logging
import sys
import signal

logger = logging.getLogger(__name__)


class SingleLevelFilter(logging.Filter):
    """Reject or approve one exact debug level. """

    def __init__(self, levelno, reject):
        """
        Args:
            levelno (int): which level do we work with (logging.DEBUG, etc.)
            reject (bool): remove messages of this level if true, remove all others if false
        """
        logging.Filter.__init__(self)
        self.levelno = levelno
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return record.levelno != self.levelno
        else:
            return record.levelno == self.levelno


def init_logging(log_filename, verbose, quiet):
    """Set up logging with default parameters:
    * default console logging level is INFO
    * ERROR, WARNING and CRITICAL are redirected to stderr

    Args:
        log_filename (str): if set, will write DEBUG log there
        verbose (bool): DEBUG level in console, overrides 'quiet'
        quiet (bool): WARNING level in console
    """
    # TODO: consider making one verbosity parameter instead of two mutually exclusive
    # TODO: default values for parameters
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)

    # add file handler if needed
    if log_filename:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(logging.DEBUG)
        # TODO: initialize all formatters in the beginning of this function
        file_handler.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)d\t%(message)s"
            ))
        logger.addHandler(file_handler)

    # console stdout and stderr handlers
    console_handler = logging.StreamHandler(sys.stdout)
    stderr_hdl = logging.StreamHandler(sys.stderr)

    # formatters
    fmt_verbose = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)d\t%(message)s",
        datefmt='%Y-%m-%d,%H:%M:%S.%f'
    )
    fmt_regular = logging.Formatter(
        "%(asctime)s [%(levelname).4s] [%(filename).8s] %(message)s", "%H:%M:%S")

    # set formatters and log levels
    if verbose:
        console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(fmt_verbose)
        stderr_hdl.setFormatter(fmt_verbose)
    elif quiet:
        console_handler.setLevel(logging.WARNING)
        console_handler.setFormatter(fmt_regular)
        stderr_hdl.setFormatter(fmt_regular)
    else:
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(fmt_regular)
        stderr_hdl.setFormatter(fmt_regular)

    # TODO: do we really need these to be redirected?
    # redirect ERROR, WARNING and CRITICAL to sterr
    f_err = SingleLevelFilter(logging.ERROR, True)
    f_warn = SingleLevelFilter(logging.WARNING, True)
    f_crit = SingleLevelFilter(logging.CRITICAL, True)
    console_handler.addFilter(f_err)
    console_handler.addFilter(f_warn)
    console_handler.addFilter(f_crit)
    logger.addHandler(console_handler)

    f_info = SingleLevelFilter(logging.INFO, True)
    f_debug = SingleLevelFilter(logging.DEBUG, True)
    stderr_hdl.addFilter(f_info)
    stderr_hdl.addFilter(f_debug)
    logger.addHandler(stderr_hdl)


# TODO: describe what are these for, how to use
def signal_handler(sig, frame):
    """ required for non-tty python runs to interrupt """
    logger.warning("Got signal %s, going to stop", sig)
    raise KeyboardInterrupt()


def ignore_handler(sig, frame):
    logger.warning("Got signal %s, ignoring", sig)


def set_sig_handler():
    uncatchable = ['SIG_DFL', 'SIGSTOP', 'SIGKILL']
    ignore = ['SIGCHLD', 'SIGCLD']
    all_sig = [s for s in dir(signal) if s.startswith("SIG")]
    for sig_name in ignore:
        try:
            sig_num = getattr(signal, sig_name)
            signal.signal(sig_num, ignore_handler)
        except Exception:
            pass
    for sig_name in [s for s in all_sig if s not in (uncatchable + ignore)]:
        try:
            sig_num = getattr(signal, sig_name)
            signal.signal(sig_num, signal_handler)
        except Exception as ex:
            logger.error("Can't set handler for %s, %s", sig_name, ex)



import pytest
from yandextank.common.monitoring import monitoring_data


@pytest.mark.parametrize(
    'metrics, result',
    [
        (
            {1: {'sens1': 1, 'sens2': 2}},
            {'timestamp': 1, 'data': {'test': {'comment': '', 'metrics': {'custom:sens1': 1, 'custom:sens2': 2}}}},
        )
    ],
)
def test_monitoring_data(metrics, result):
    assert monitoring_data('test', metrics, '') == result



import datetime
import logging
import subprocess
import time
import os
import re
import shutil
from threading import Event

import yaml

from yandextank.contrib.netort.netort.resource import TempDownloaderOpenerProtocol

from .reader import PandoraStatsReader
from ..Console import Plugin as ConsolePlugin
from ..Console import screen as ConsoleScreen
from ..Phantom import PhantomReader, string_to_df
from ...common.interfaces import AbstractInfoWidget, GeneratorPlugin
from ...common.util import tail_lines, FileMultiReader

logger = logging.getLogger(__name__)


class Plugin(GeneratorPlugin):
    """    Pandora load generator plugin    """

    OPTION_CONFIG = "config"
    SECTION = "pandora"
    DEFAULT_REPORT_FILE = "phout.log"
    DEFAULT_EXPVAR_PORT = 1234

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.output_finished = Event()
        self.enum_ammo = False
        self.pandora_cmd = None
        self.pandora_config_file = None
        self.config_contents = None
        self.custom_config = False
        self.expvar = self.get_option('expvar')
        self.expvar_enabled = self.expvar
        self.expvar_port = self.DEFAULT_EXPVAR_PORT
        self.report_files = None
        self.__address = None
        self.__schedule = None
        self.ammofile = None
        self.process_stderr_file = None
        self.resources = []

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = [
            "pandora_cmd", "buffered_seconds",
            "config_content", "config_file"
        ]
        return opts

    def configure(self):
        self.report_files = [self.get_option("report_file")]
        self.buffered_seconds = self.get_option("buffered_seconds")
        self.affinity = self.get_option("affinity", "")
        self.resources = self.get_option("resources")

        # if we use custom pandora binary, we can download it and make it executable
        self.pandora_cmd = self.get_resource(self.get_option("pandora_cmd"), "./pandora", permissions=0o755)

        # download all resources from self.get_options("resources")
        if len(self.resources) > 0:
            for resource in self.resources:
                self.get_resource(resource["src"], resource["dst"])

        # get config_contents and patch it: expand resources via resource manager
        # config_content option has more priority over config_file
        if self.get_option("config_content"):
            logger.info('Found config_content option configuration')
            self.config_contents = self.__patch_raw_config_and_dump(self.get_option("config_content"))
        elif self.get_option("config_file"):
            logger.info('Found config_file option configuration')
            with open(self.get_option("config_file"), 'rb') as config:
                external_file_config_contents = yaml.load(config.read(), Loader=yaml.FullLoader)
            self.config_contents = self.__patch_raw_config_and_dump(external_file_config_contents)
        else:
            raise RuntimeError("Neither pandora.config_content, nor pandora.config_file specified")
        logger.debug('Config after parsing for patching: %s', self.config_contents)

        # find report filename and add to artifacts
        self.report_files = [pool['result']['destination'] for pool in self.config_contents['pools']]
        for f in self.report_files:
            with open(f, 'w'):
                pass
            self.core.add_artifact_file(f)

    def __patch_raw_config_and_dump(self, cfg_dict):
        if not cfg_dict:
            raise RuntimeError('Empty pandora config')
        # patch
        config_content = self.patch_config(cfg_dict)
        # dump
        self.pandora_config_file = self.core.mkstemp(".yaml", "pandora_config_")
        self.core.add_artifact_file(self.pandora_config_file)
        with open(self.pandora_config_file, 'w') as config_file:
            yaml.dump(config_content, config_file)
        return config_content

    def patch_config(self, config):
        """
        download remote resources, replace links with local filenames
        add result file section
        :param dict config: pandora config
        """
        # get expvar parameters
        if config.get("monitoring"):
            if config["monitoring"].get("expvar"):
                self.expvar_enabled = config["monitoring"]["expvar"].get("enabled")
                if config["monitoring"]["expvar"].get("port"):
                    self.expvar_port = config["monitoring"]["expvar"].get("port")
        # or set if expvar not exists
        elif not self.expvar:
            config["monitoring"] = {
                "expvar": {
                    "enabled": True,
                }
            }
            self.expvar_enabled = True

        # FIXME this is broken for custom ammo providers due to interface incompatibility
        # FIXME refactor pandora plx
        for n, pool in enumerate(config['pools']):
            if pool.get('ammo', {}).get('file', ''):
                self.ammofile = pool['ammo']['file']
                opener = self.core.resource_manager.get_opener(self.ammofile)
                if isinstance(opener, TempDownloaderOpenerProtocol):
                    pool['ammo']['file'] = opener.download_file(True, try_ungzip=True)
                else:
                    pool['ammo']['file'] = opener.get_filename

            if not pool.get('result') or 'phout' not in pool.get('result', {}).get('type', ''):
                logger.warning('Seems like pandora result file not specified... adding defaults')
                pool['result'] = dict(
                    destination=f"{n}_{self.DEFAULT_REPORT_FILE}",
                    type='phout',
                )

            if pool.get('gun').get('answlog'):
                answ = pool.get('gun').get('answlog')
                if answ.get('enabled') is True:
                    if not answ.get('path'):
                        answ_path = self.core.mkstemp(".log", "answ_")
                        logger.warning('Seems like pandora answer log file not specified... adding default path: %s', answ_path)
                        self.core.add_artifact_file(answ_path)
                        config['pools'][n]['gun']['answlog']['path'] = answ_path
                    else:
                        answ_path = answ.get('path')
                        self.core.add_artifact_file(answ_path)

        return config

    @property
    def address(self):
        if not self.__address:
            for pool in self.config_contents['pools']:
                if pool.get('gun', {}).get('target'):
                    self.__address = pool.get('gun', {}).get('target')
                    break
            else:
                self.__address = 'unknown'
        return self.__address

    @property
    def schedule(self):
        if not self.__schedule:
            for pool in self.config_contents['pools']:
                if pool.get('rps'):
                    self.__schedule = pool.get('rps')
                    break
            else:
                self.__schedule = 'unknown'
        return self.__schedule

    def get_info(self):
        return self.Info(
            address=self.address,
            ammo_file=self.ammofile,
            duration=0,
            instances=0,
            loop_count=0,
            port=self.address.split(':')[-1],
            rps_schedule=self.schedule
        )

    def get_reader(self, parser=string_to_df):
        if self.reader is None:
            self.reader = [FileMultiReader(f, self.output_finished) for f in self.report_files]
        return [PhantomReader(reader.get_file(), parser=parser) for reader in self.reader]

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = PandoraStatsReader(self.expvar_enabled, self.expvar_port)
        return self.stats_reader

    def get_resource(self, resource, dst, permissions=0o644):
        opener = self.core.resource_manager.get_opener(resource)
        if isinstance(opener, TempDownloaderOpenerProtocol):
            tmp_path = opener.download_file(True, try_ungzip=True)
            shutil.copy(tmp_path, dst)
            logger.info('Successfully moved resource %s', dst)
        else:
            dst = opener.get_filename
        try:
            os.chmod(dst, permissions)
        except OSError:
            logger.warning('Cannot change permissions to %s', dst)
        return dst

    def prepare_test(self):
        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
            console = None

        if console:
            widget = PandoraInfoWidget(self)
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        args = [self.pandora_cmd] +\
            (['-expvar'] if self.expvar else []) +\
            [self.pandora_config_file]
        if self.affinity:
            self.core.__setup_affinity(self.affinity, args=args)
        logger.info("Starting: %s", args)
        self.start_time = time.time()
        self.process_stderr_file = self.core.mkstemp(".log", "pandora_")
        self.core.add_artifact_file(self.process_stderr_file)
        self.process_stderr = open(self.process_stderr_file, 'w')
        try:
            self.process = subprocess.Popen(
                args,
                stderr=self.process_stderr,
                stdout=self.process_stderr,
                close_fds=True)
        except OSError:
            logger.debug(
                "Unable to start Pandora binary. Args: %s", args, exc_info=True)
            raise RuntimeError(
                "Unable to start Pandora binary and/or file does not exist: %s" % args)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None and retcode == 0:
            logger.info("Pandora subprocess done its work successfully and finished w/ retcode 0")
            self.output_finished.set()
            return retcode
        elif retcode is not None and retcode != 0:
            lines_amount = 20
            logger.info("Pandora finished with retcode %s. Last %s logs of Pandora log:", retcode, lines_amount)
            self.output_finished.set()
            last_log_contents = tail_lines(self.process_stderr_file, lines_amount)
            for logline in last_log_contents:
                logger.info(logline.strip('\n'))
            with open(self.process_stderr_file) as stderr:
                for line in stderr:
                    if self.check_log_line_contains_error(line):
                        self.errors.append(line.strip())
            return abs(retcode)
        else:
            return -1

    @staticmethod
    def check_log_line_contains_error(line):
        return re.search('^panic:|ERROR|FATAL', line) is not None

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.warning("Terminating worker process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stderr:
                self.process_stderr.close()
        else:
            logger.info("Seems Pandora subprocess finished")
        self.output_finished.set()
        return retcode


class PandoraInfoWidget(AbstractInfoWidget):
    ''' Right panel widget '''

    def __init__(self, pandora):
        AbstractInfoWidget.__init__(self)
        self.krutilka = ConsoleScreen.krutilka()
        self.owner = pandora
        self.reqps = 0
        self.active = 0

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stats):
        self.reqps = stats["metrics"]["reqps"]
        self.active = stats["metrics"]["instances"]

    def render(self, screen):
        text = " Pandora Test %s" % next(self.krutilka)
        space = screen.right_panel_width - len(text) - 1
        left_spaces = space // 2
        right_spaces = space // 2

        dur_seconds = int(time.time()) - int(self.owner.start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_BROWN + '~' * left_spaces + \
            text + ' ' + '~' * right_spaces + screen.markup.RESET + "\n"
        template += "Command Line: %s\n"
        template += "    Duration: %s\n"
        template += "  Requests/s: %s\n"
        template += " Active reqs: %s\n"
        template += "      Target: %s\n"
        template += "    Schedule: \n%s\n"
        data = (
            self.owner.pandora_cmd,
            duration,
            self.reqps,
            self.active,
            self.owner.address,
            yaml.dump(self.owner.schedule)
        )

        return template % data



import os.path
import getpass
import logging
import tempfile
import pkg_resources
from copy import deepcopy
from ..Telegraf.decoder import decoder
from .config_parser import parse_xml, parse_yaml, ParseError, TARGET_HINT_PLACEHOLDER
from yandextank.common.util import read_resource
import configparser
from requests.structures import CaseInsensitiveDict

logger = logging.getLogger(__name__)


class ConfigManager(object):
    """
    Config reader and parser helper.
    XML support
    """

    @staticmethod
    def parse_config(config):
        parsers = [
            ('xml', parse_xml),
            ('yaml', parse_yaml)
        ]
        for name, parser in parsers:
            try:
                return parser(config)
            except ParseError as exc:
                logger.warning("%s config parsing error: %s", name, exc)

        logger.warning("Couldn't parse monitoring config at %s. Using default config", config)
        return []

    @staticmethod
    def _apply_defaults(host_config, defaults: dict):
        if not defaults:
            return

        for k, v in defaults.items():
            if hasattr(host_config, k):
                if getattr(host_config, k) is None:
                    setattr(host_config, k, deepcopy(v))
            elif not host_config.get(k):
                host_config[k] = deepcopy(v)

    def getconfig(self, filename, target_hint, defaults: dict = None):
        """Prepare config data."""
        defaults = defaults or {}
        try:
            config = read_resource(filename)
        except IOError:
            logger.exception("Error loading config")
            raise RuntimeError("Can't read monitoring config %s" % filename)

        hosts = self.parse_config(config)
        config = []
        for host in hosts:
            host_config = self.get_host_config(host, target_hint)
            self._apply_defaults(host_config, defaults)
            config.append(host_config)
        return config

    def get_host_config(self, host, target_hint):
        defaults = {
            "cpu": {
                "name": '[inputs.cpu]',
                "percpu": 'false',
                "fielddrop": '["time_*", "usage_guest_nice"]'
            },
            "memory": {
                "name": '[inputs.mem]',
                "fielddrop":
                '["active", "inactive", "total", "used_per*", "avail*"]',
            },
            "disk": {
                "name": '[inputs.diskio]',
                "devices": '[{devices}]'.format(
                    devices=",".join(
                        ['"vda%s","sda%s"' % (num, num) for num in range(6)])),
            },
            "net": {
                "name": '[inputs.net]',
                "interfaces": '[{interfaces}]'.format(
                    interfaces=",".join(
                        ['"eth%s"' % (num) for num in range(6)])),
                "fielddrop":
                '["icmp*", "ip*", "udplite*", "tcp*", "udp*", "drop*", "err*"]',
            },
            "nstat": {
                "name": '[inputs.nstat]',
                "fieldpass": '["TcpRetransSegs"]',
            },
            "netstat": {
                "name": '[inputs.netstat]',
            },
            "netresponse": {
                "name": '[inputs.net_response]',
                "protocol": '"tcp"',
                "address": '":80"',
                "timeout": '"1s"'
            },
            "system": {
                "name": '[inputs.system]',
                "fielddrop": '["n_users", "n_cpus", "uptime*"]',
            },
            "kernel": {
                "name": '[inputs.kernel]',
                "fielddrop": '["boot_time"]',
            },
            "kernelvmstat": {
                "name": '[inputs.kernel_vmstat]',
                "fieldpass": '["pgfault", "pgmajfault"]',
            }
        }

        # compatibility with native telegraf metric names
        defaults['mem'] = defaults['memory']
        defaults['diskio'] = defaults['disk']
        defaults['net_response'] = defaults['netresponse']
        defaults['kernel_vmstat'] = defaults['kernelvmstat']

        defaults_enabled = ['cpu', 'memory', 'disk', 'net', 'system', 'kernel']
        defaults_boolean = [
            'percpu', 'round_interval', 'fielddrop', 'fieldpass', 'interfaces',
            'devices'
        ]
        hostname = host.address.lower()
        if hostname == TARGET_HINT_PLACEHOLDER:
            if not target_hint:
                raise ValueError(f"Can't use `{TARGET_HINT_PLACEHOLDER}` keyword with no target parameter specified")
            logger.debug("Using target hint: %s", target_hint)
            hostname = target_hint.lower()
        custom = []
        startups = []
        shutdowns = []
        sources = []
        telegrafraw = []
        # agent defaults
        host_config = CaseInsensitiveDict()
        for metric in host.metrics:
            metric_name = str(metric.name).lower()
            if metric_name in defaults:
                for key in tuple(defaults[metric_name].keys()):
                    if key != 'name' and key not in defaults_boolean:
                        value = metric.get(key, None)
                        if value:
                            defaults[metric_name][key] = "'{value}'".format(
                                value=value)
                    elif key in defaults_boolean:
                        value = metric.get(key, None)
                        if value:
                            defaults[metric_name][key] = "{value}".format(
                                value=value)
                host_config[metric_name] = defaults[metric_name]
            # custom metrics
            if metric_name == 'custom':
                isdiff = metric.get('diff', 0)
                cmd = {
                    'cmd': metric.text,
                    'label': metric.get('label'),
                    'diff': isdiff
                }
                custom.append(cmd)
            elif metric_name == 'startup':
                startups.append(metric.text)
            elif metric_name == 'shutdown':
                shutdowns.append(metric.text)
            elif metric_name == 'source':
                sources.append(metric.text)
            elif metric_name == 'telegrafraw':
                telegrafraw.append(metric.text)
        if len(host_config) == 0:
            logger.info('Empty host config, using defaults')
            for section in defaults_enabled:
                host_config[section] = defaults[section]
        result = {
            'host_config': host_config,
            'port': int(host.get('port', 22)),
            'python': host.get('python', '/usr/bin/env python3'),
            'interval': host.get('interval', 1),
            'username': host.get('username', getpass.getuser()),
            'telegraf': host.get('telegraf', '/usr/bin/telegraf'),
            'comment': host.get('comment', ''),
            'ssh_key_path': host.get('ssh_key_path'),
            'custom': custom,
            'host': hostname,
            'startup': startups,
            'shutdown': shutdowns,
            'source': sources,
            'telegrafraw': telegrafraw
        }
        logger.info("Telegraf Result config %s", result)
        return result


class AgentConfig(object):
    """ Agent config generator helper """

    def __init__(self, config, old_style_configs):
        self.host = config['host']
        self.custom = config['custom']
        self.startups = config['startup']
        self.shutdowns = config['shutdown']
        self.sources = config['source']
        self.interval = config['interval']
        self.comment = config['comment']
        self.telegrafraw = config['telegrafraw']
        self.host_config = config['host_config']
        self.old_style_configs = old_style_configs

    def create_startup_config(self):
        """ Startup and shutdown commands config
        Used by agent.py on the target

        """
        cfg_path = "agent_startup_{}.cfg".format(self.host)
        if os.path.isfile(cfg_path):
            logger.info(
                'Found agent startup config file in working directory with the same name as created for host %s.\n'
                'Creating new one via tempfile. This will affect predictable filenames for agent artefacts',
                self.host)
            handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_')
            os.close(handle)
        try:
            config = configparser.RawConfigParser(strict=False)
            # FIXME incinerate such a string formatting inside a method call
            # T_T
            config.add_section('startup')
            [
                config.set('startup', "cmd%s" % idx, cmd)
                for idx, cmd in enumerate(self.startups)
            ]
            config.add_section('shutdown')
            [
                config.set('shutdown', "cmd%s" % idx, cmd)
                for idx, cmd in enumerate(self.shutdowns)
            ]
            config.add_section('source')
            [
                config.set('source', "file%s" % idx, path)
                for idx, path in enumerate(self.sources)
            ]
            with open(cfg_path, 'w') as fds:
                config.write(fds)

        except Exception:
            logger.exception('Error trying to create monitoring startups config. Malformed?')
        return cfg_path

    def create_custom_exec_script(self):
        """ bash script w/ custom commands inside
        inspired by half a night trying to avoid escaping bash special characters

        """
        cfg_path = "agent_customs_{}.cfg".format(self.host)
        if os.path.isfile(cfg_path):
            logger.info(
                'Found agent custom execs config file in working directory with the same name as created for host %s.\n'
                'Creating new one via tempfile. This will affect predictable filenames for agent artefacts',
                self.host)
            handle, cfg_path = tempfile.mkstemp('.sh', 'agent_customs_')
            os.close(handle)

        cmds = ""
        for idx, cmd in enumerate(self.custom):
            cmds += "-{idx}) {cmd};;\n".format(idx=idx, cmd=cmd['cmd'])
        customs_script = """
        #!/bin/sh
        while :
        do
            case "$1" in
            {cmds}
            *) break;;
            esac
            shift
        done
        """.format(cmds=cmds)

        with open(cfg_path, 'w') as fds:
            fds.write(customs_script)
        return cfg_path

    def create_collector_config(self, workdir):
        """ Telegraf collector config,
        toml format

        """
        cfg_path = "agent_collector_{}.cfg".format(self.host)
        if os.path.isfile(cfg_path):
            logger.info(
                'Found agent config file in working directory with the same name as created for host %s.\n'
                'Creating new one via tempfile. This will affect predictable filenames for agent artefacts',
                self.host)
            handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_collector_')
            os.close(handle)

        self.monitoring_data_output = "{remote_folder}/monitoring.rawdata".format(
            remote_folder=workdir)

        defaults_old_enabled = ['cpu', 'memory', 'disk', 'net', 'system']

        try:
            config = configparser.RawConfigParser(strict=False)

            config.add_section("global_tags")
            config.add_section("agent")
            config.set(
                "agent",
                "interval",
                "'{interval}s'".format(interval=self.interval))
            config.set("agent", "round_interval", "true")
            config.set("agent", "flush_interval", "'1s'")
            config.set("agent", "collection_jitter", "'0s'")
            config.set("agent", "flush_jitter", "'1s'")

            for section in self.host_config.keys():
                # telegraf-style config
                if not self.old_style_configs:
                    config.add_section(
                        "{section_name}".format(
                            section_name=self.host_config[section]['name']))
                    for key, value in self.host_config[section].items():
                        if key != 'name':
                            config.set(
                                "{section_name}".format(
                                    section_name=self.host_config[section][
                                        'name']),
                                "{key}".format(key=key),
                                "{value}".format(value=value))
                # monitoring-style config
                else:
                    if section in defaults_old_enabled:
                        config.add_section(
                            "{section_name}".format(
                                section_name=self.host_config[section]['name']))
                        for key, value in self.host_config[section].items():
                            if key in [
                                    'fielddrop', 'fieldpass', 'percpu',
                                    'devices', 'interfaces'
                            ]:
                                config.set(
                                    "{section_name}".format(
                                        section_name=self.host_config[section][
                                            'name']),
                                    "{key}".format(key=key),
                                    "{value}".format(value=value))

            # outputs
            config.add_section("[outputs.file]")
            config.set(
                "[outputs.file]",
                "files",
                "['{config}']".format(config=self.monitoring_data_output))
            config.set("[outputs.file]", "data_format", "'json'")

            with open(cfg_path, 'w') as fds:
                config.write(fds)

            # dirty hack, this allow to avoid bash escape quoting, we're pushing shell script w/ arguments
            # index of argument is index of custom metric in our config
            inputs = ""
            for idx, cmd in enumerate(self.custom):
                inputs += "[[inputs.exec]]\n"
                inputs += "commands = ['/bin/sh {workdir}/agent_customs.sh -{idx}']\n".format(
                    workdir=workdir, idx=idx)
                inputs += "data_format = 'value'\n"
                inputs += "data_type = 'float'\n"
                inputs += "name_prefix = '{}_'\n\n".format(cmd.get('label'))
                if cmd['diff']:
                    decoder.diff_metrics['custom'].append(
                        decoder.find_common_names(cmd.get('label')))

            with open(cfg_path, 'a') as fds:
                fds.write(inputs)

            # telegraf raw configuration into xml
            telegraf_raw = "".join(self.telegrafraw)

            with open(cfg_path, 'a') as fds:
                fds.write(telegraf_raw)

        except Exception:
            logger.exception('Error trying to create monitoring config. Malformed?')
        return cfg_path


def create_agent_py(agent_filename):
    with open(agent_filename, 'w') as f:
        f.write(read_resource(pkg_resources.resource_filename('yandextank.plugins.Telegraf', 'agent/agent.py')))
    os.chmod(agent_filename, 0o775)
    return os.path.abspath(agent_filename)



import os
import yaml
from xml.etree import ElementTree
from requests.structures import CaseInsensitiveDict

from typing import List

TARGET_HINT_PLACEHOLDER = '[target]'
YAML_HOSTS_SECTION = 'hosts'
YAML_METRICS_SECTION = 'metrics'
YAML_CUSTOM_METRIC = 'custom'
YAML_CUSTOM_METRIC_CMD = 'cmd'
XML_HOST_TAG_NAME = 'Host'
XML_HOST_ADDRESS_ATTR = 'address'


class ParseError(Exception):
    pass


class Metric(CaseInsensitiveDict):
    name: str
    text: str

    def __init__(self, name, text, data):
        super().__init__(data)
        self.name = name
        self.text = text


class Host(CaseInsensitiveDict):
    address: str
    metrics: List[Metric]

    def __init__(self, address, metrics, data):
        super().__init__(data)
        self.address = address
        self.metrics = metrics


def parse_xml(config) -> List[Host]:
    try:
        if os.path.exists(config):
            tree = ElementTree.parse(config)
        else:
            tree = ElementTree.fromstring(config)
    except ElementTree.ParseError as ex:
        raise ParseError(ex)

    result = []
    hosts = tree.findall(XML_HOST_TAG_NAME)
    for host in hosts:
        hostname = host.get(XML_HOST_ADDRESS_ATTR, '').lower()
        metrics = [Metric(m.tag, m.text, m.attrib) for m in host]
        result.append(Host(hostname, metrics, host.attrib))
    return result


def parse_yaml(config) -> List[Host]:
    try:
        if os.path.exists(config):
            with open(config, "r") as stream:
                yaml_content = yaml.safe_load(stream)
        else:
            yaml_content = yaml.safe_load(config)
    except yaml.YAMLError as ex:
        raise ParseError(ex)

    result = []

    yaml_content = yaml_content or {}

    global_inputs = yaml_content.get(YAML_METRICS_SECTION, {})
    agents = yaml_content.get(YAML_HOSTS_SECTION, {})

    # if no "agents:" provided use default host
    if len(agents) == 0:
        agents[TARGET_HINT_PLACEHOLDER] = None
    for hostname, hostdata in agents.items():
        metrics = []
        local_inputs = global_inputs.copy()
        hostdata = hostdata or {}

        local_inputs.update(hostdata.get(YAML_METRICS_SECTION, {}))

        for mname, mdata in local_inputs.items():
            if mdata is None:
                mdata = ''

            if mname.lower() == YAML_CUSTOM_METRIC and isinstance(mdata, dict):
                mtext = mdata.get(YAML_CUSTOM_METRIC_CMD, '') or str(mdata)
            else:
                mtext = str(mdata)
            metrics.append(Metric(mname, mtext, mdata if isinstance(mdata, dict) else {}))

        result.append(Host(hostname, metrics, hostdata if isinstance(hostdata, dict) else {}))
    return result



import errno
import collections
import logging
import queue
import subprocess
import time
import re
import pandas as pd
from threading import Event
from typing import Optional

from ...common.interfaces import AbstractPlugin, GeneratorPlugin, AggregateResultListener, AbstractInfoWidget, \
    StatsReader
from ...common.util import FileMultiReader, FileScanner, tail_lines
from ..Console import Plugin as ConsolePlugin
from ..Phantom import PhantomReader
from yandextank.aggregator import TimeChopper
from yandextank.aggregator.aggregator import DataPoller


_INFO = collections.namedtuple(
    "Info",
    "address, port, instances, ammo_count, loop_count, duration, steps, stat_log, rps_schedule, ammo_file"
)
_LOGGER = logging.getLogger(__name__)

_PROCESS_KILL_TIMEOUT = 10  # Kill running process after specified number of seconds


class Plugin(GeneratorPlugin):
    """Simple executor of shooting process with phantom compatible output"""

    SECTION = 'shootexec'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.output_finished = Event()
        self.stats_reader = None
        self.reader = None
        self._stderr_path = None
        self.__process = None
        self.__stderr_file = None
        self.__processed_ammo_count = 0
        self.start_time = 0
        self.opened_file = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return ["cmd", "output_path", "stats_path"]

    def configure(self):
        self.__cmd = self.get_option("cmd")

        self.__output_path = self.get_option("output_path")
        self.core.add_artifact_file(self.__output_path)

        self.__stats_path = self.get_option("stats_path")
        if self.__stats_path:
            self.core.add_artifact_file(self.__stats_path)

    def get_reader(self):
        if self.reader is None:
            # Touch output_path to clear it
            open(self.__output_path, "w").close()
            self.file_reader = FileMultiReader(self.__output_path, self.output_finished)
            self.add_cleanup(lambda: self.file_reader.close)
            self.reader = PhantomReader(self.file_reader.get_file())
            if not self.__stats_path:
                self.reader = _ShootExecReader(self.reader, self.core.data_poller)
        return self.reader

    def get_stats_reader(self):
        if self.stats_reader is None:
            if self.__stats_path:
                open(self.__stats_path, 'w').close()
                self.stats_reader = _FileStatsReader(self.__stats_path)
            else:
                reader = self.get_reader()
                if hasattr(reader, 'stats_reader'):
                    self.stats_reader = reader.stats_reader
                else:
                    self.stats_reader = _DummyStatsReader()
        return self.stats_reader

    def prepare_test(self):
        self._stderr_path = self.core.mkstemp(".log", "shootexec_stdout_stderr_")
        self.__stderr_file = open(self._stderr_path, 'w')

        _LOGGER.debug("Linking sample reader to aggregator. Reading samples from %s", self.__output_path)

        self.start_time = time.time()
        self.core.job.aggregator.add_result_listener(self)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            _LOGGER.debug(ex)
            console = None

        if console:
            widget = _InfoWidget(self)
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        _LOGGER.info("Starting shooting process: '%s'", self.__cmd)
        self.__process = subprocess.Popen(
            self.__cmd,
            shell=True,
            stderr=self.__stderr_file,
            stdout=self.__stderr_file,
            close_fds=True
        )

        _LOGGER.info("Shooting proces is ready to use")

    def is_test_finished(self):
        retcode = self.__process.poll()
        if retcode is not None:
            _LOGGER.info("Shooting process done its work with exit code: %s", retcode)
            if retcode != 0:
                lines_amount = 20
                error, lines = self.extract_error_from_log(lines_amount)
                if len(lines) > 0:
                    if error is None:
                        error = 'Last {} of ShootExec generator log: {}'.format(lines_amount, '\n'.join(lines))
                    _LOGGER.info("Last %s logs of ShootExec generator log: %s", lines_amount, '\n'.join(lines))
                if error is None:
                    error = 'Unknown generator error'
                self.errors.append(error)
            self.output_finished.set()
            return abs(retcode)
        else:
            return -1

    def extract_error_from_log(self, lines_amount=20):
        last_log_contents = tail_lines(self._stderr_path, lines_amount)
        lines = []
        error = None
        for logline in last_log_contents:
            line = logline.strip('\n')
            lines.append(line)
            if self.check_log_line_contains_error(line):
                error = line
        return error, lines

    @staticmethod
    def check_log_line_contains_error(line):
        return re.search('^panic:|ERROR|FATAL', line) is not None

    def end_test(self, retcode):
        if self.__process and self.__process.poll() is None:
            _LOGGER.warning("Terminating shooting process with PID %s", self.__process.pid)
            self.__terminate()
        else:
            _LOGGER.info("Seems shooting process finished OK")
        self.output_finished.set()
        return retcode

    def post_process(self, retcode):
        return retcode

    def on_aggregated_data(self, data, stats):
        self.__processed_ammo_count += data["overall"]["interval_real"]["len"]
        _LOGGER.debug("Processed ammo count: %s", self.__processed_ammo_count)

    def get_info(self):
        """ returns info object """
        return _INFO(
            "",
            "",
            "0",
            "0",
            "0",
            time.time() - self.start_time,
            None,
            "",
            "",
            ""
        )

    def __terminate(self):
        """Gracefull termination of running process"""

        if self.__stderr_file:
            self.__stderr_file.close()

        if not self.__process:
            return

        waitfor = time.time() + _PROCESS_KILL_TIMEOUT
        while time.time() < waitfor:
            try:
                self.__process.terminate()
            except EnvironmentError as e:
                if e.errno != errno.ESRCH:
                    _LOGGER.warning("Failed to terminate process '{}': {}".format(self.__cmd, e))
                return
            time.sleep(0.1)

        try:
            self.__process.kill()
        except EnvironmentError as e:
            if e.errno != errno.ESRCH:
                _LOGGER.warning("Failed to kill process '{}': {}".format(self.__cmd, e))
            return


class _InfoWidget(AbstractInfoWidget, AggregateResultListener):
    """
    Widget with information about current run state
    """

    def get_index(self):
        return 2

    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.owner = sender

    def render(self, screen):
        return ""

    def on_aggregated_data(self, data, stats):
        pass


class _FileStatsReader(FileScanner, StatsReader):
    """
    Read shooting stats line by line

    Line format is 'timestamp\trps\tinstances'
    """

    def __init__(self, *args, **kwargs):
        super(_FileStatsReader, self).__init__(*args, **kwargs)
        self.__last_ts = 0

    def _read_data(self, lines):
        """
        Parse lines and return stats
        """

        results = []
        for line in lines:
            timestamp, rps, instances = line.split("\t")
            curr_ts = int(float(timestamp))  # We allow floats here, but tank expects only seconds
            if self.__last_ts < curr_ts:
                self.__last_ts = curr_ts
                results.append(self.stats_item(self.__last_ts, float(rps), float(instances)))
        return results


class _ShootExecReader(object):
    def __init__(self, phout_reader: PhantomReader, poller: DataPoller):
        self._inner_reader = phout_reader
        self.closed = False
        self.stat_queue = queue.Queue()
        self.stats_reader = _ShootExecStatAggregator(
            TimeChopper([poller.poll(self._read_stat_queue())]))

    @property
    def buffer(self):
        return self._inner_reader.buffer

    def __iter__(self):
        return self

    def __next__(self):
        try:
            res = self._inner_reader.__next__()
            if res is not None:
                self.stat_queue.put(res)
            return res
        except StopIteration:
            self.closed = True
            raise

    def _read_stat_queue(self):
        while True:
            try:
                yield self._prepare_stat_item(self.stat_queue.get_nowait())
            except queue.Empty:
                if self.closed:
                    return None
                yield None

    def _prepare_stat_item(self, item: Optional[pd.DataFrame]):
        if item is None:
            return None
        item = item.copy(deep=True)
        item['send_ts'] = item['send_ts'].astype(int)
        item.set_index(['send_ts'], inplace=True)
        return item


class _ShootExecStatAggregator(object):
    def __init__(self, source):
        self.source = source

    def __iter__(self):
        for ts, _, rps in self.source:
            yield [{
                'ts': ts,
                'metrics': {
                    'instances': 0,
                    'reqps': rps
                }
            }]

    def close(self):
        pass


class _DummyStatsReader(StatsReader):
    """
    Dummy stats reader for shooters without stats file
    """

    def __init__(self):
        self.__closed = False
        self.__last_ts = 0

    def __iter__(self):
        while not self.__closed:
            cur_ts = int(time.time())
            if cur_ts > self.__last_ts:
                yield [self.stats_item(cur_ts, 0, 0)]
                self.__last_ts = cur_ts
            else:
                yield []

    def close(self):
        self.__closed = True



import logging
from typing import Optional
from yandextank.contrib.netort.netort.resource import ResourceManager, manager

from . import info
from . import instance_plan as ip
from . import load_plan as lp
from . import missile
from .mark import get_marker
from .module_exceptions import StepperConfigurationError, AmmoFileError


class ComponentFactory():
    def __init__(
            self,
            rps_schedule=None,
            http_ver='1.1',
            ammo_file=None,
            instances_schedule=None,
            instances=1000,
            loop_limit=-1,
            ammo_limit=-1,
            uris=None,
            headers=None,
            autocases=0,
            enum_ammo=False,
            ammo_type='phantom',
            chosen_cases=None,
            use_cache=True,
            resource_manager: Optional[ResourceManager] = None,
    ):
        self.log = logging.getLogger(__name__)
        self.ammo_file = ammo_file
        self._ammo_type_unchecked = ammo_type
        self.rps_schedule = rps_schedule
        self.http_ver = http_ver
        self.instances_schedule = instances_schedule
        loop_limit = int(loop_limit)
        if loop_limit == -1:  # -1 means infinite
            loop_limit = None
        ammo_limit = int(ammo_limit)
        if ammo_limit == -1:  # -1 means infinite
            ammo_limit = None
        if loop_limit is None and ammo_limit is None and not rps_schedule:
            # we should have only one loop if we have instance_schedule
            loop_limit = 1
        info.status.loop_limit = loop_limit
        info.status.ammo_limit = ammo_limit
        info.status.publish("instances", instances)
        self.uris = uris
        if self.uris and loop_limit:
            info.status.ammo_limit = len(self.uris) * loop_limit
        self.headers = headers if headers is not None else []
        self.marker = get_marker(autocases, enum_ammo)
        self.chosen_cases = chosen_cases or []
        self.use_cache = use_cache
        self.resource_manager = resource_manager or manager

    @property
    def ammo_type(self):
        if self._ammo_type_unchecked != 'phantom':
            return self._ammo_type_unchecked

        opener = self.resource_manager.get_opener(self.ammo_file)
        with opener(self.use_cache) as ammo:
            try:
                ammo_str = next(ammo).decode('utf-8')
            except StopIteration:
                self.log.exception("Couldn't read first line of ammo file")
                raise AmmoFileError("Couldn't read first line of ammo file")

            if not ammo_str[0].isdigit():
                self.log.info(
                    "Setting ammo_type 'uri' because ammo is not started with digit and you did not specify ammo format"
                )
                return 'uri'
            self.log.info("Default ammo type ('phantom') used, use 'phantom.ammo_type' option to override it")
            return self._ammo_type_unchecked

    def get_load_plan(self):
        """
        return load plan (timestamps generator)
        """
        if self.rps_schedule and self.instances_schedule:
            raise StepperConfigurationError(
                'Both rps and instances schedules specified. You must specify only one of them'
            )
        elif self.rps_schedule:
            info.status.publish('loadscheme', self.rps_schedule)
            return lp.create(self.rps_schedule)
        elif self.instances_schedule:
            info.status.publish('loadscheme', self.instances_schedule)
            return ip.create(self.instances_schedule)
        else:
            self.instances_schedule = []
            info.status.publish('loadscheme', self.instances_schedule)
            return ip.create(self.instances_schedule)

    def _check_exactly_one_source_specified(self):
        sources = {
            'uris': self.uris,
            'ammo_file': self.ammo_file
        }
        active_sources = {k: v for k, v in sources.items() if v}
        if len(active_sources) != 1:
            self.log.error('active sources: %s', active_sources)
            raise StepperConfigurationError(
                f'Exactly one of sources {list(sources.keys())} should be specified, '
                f'but found {list(active_sources.keys())} in config.')

    def _get_ammo_generator(self):
        self._check_exactly_one_source_specified()
        if self.uris:
            return missile.UriStyleGenerator(self.uris, self.headers, http_ver=self.http_ver)

        if self.ammo_type not in missile.FILE_READERS:
            raise NotImplementedError('No such ammo type implemented: "%s"' % self.ammo_type)

        return missile.FILE_READERS[self.ammo_type](
            self.ammo_file, headers=self.headers, http_ver=self.http_ver, use_cache=self.use_cache, resource_manager=self.resource_manager)

    def get_ammo_generator(self):
        ammo_gen = self._get_ammo_generator()
        self.log.info("Using %s ammo reader" % type(ammo_gen).__name__)
        return ammo_gen

    def get_marker(self):
        return self.marker

    def get_filter(self):
        if len(self.chosen_cases):

            def is_chosen_case(ammo_tuple):
                return ammo_tuple[1] in self.chosen_cases

            return is_chosen_case
        else:
            return lambda ammo_tuple: True



# coding=utf-8
# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
# pylint: disable=method-hidden
import logging
import os
import pwd
import re
import sys
import time
import datetime
import yaml
from urllib.parse import urljoin
from abc import ABC, abstractmethod

from queue import Empty, Queue
from builtins import str
import requests
import threading

from ...common.interfaces import AbstractPlugin, \
    MonitoringDataListener, AggregateResultListener, AbstractInfoWidget
from ...common.util import expand_to_seconds
from ..Autostop import Plugin as AutostopPlugin
from ..Console import Plugin as ConsolePlugin
from .client import APIClient, OverloadClient, LPRequisites
from ...common.util import FileScanner

from yandextank.contrib.netort.netort.data_processing import Drain

logger = logging.getLogger(__name__)  # pylint: disable=C0103

LOADTESTING_CONFIG_PATH = '/etc/tank_client/config.yaml'
LOADTESTING_CONFIG_PATH_ENV = 'LOADTESTING_AGENT_CONFIG'


class BackendTypes(object):
    OVERLOAD = 'OVERLOAD'
    LUNAPARK = 'LUNAPARK'

    @classmethod
    def identify_backend(cls, api_address, cfg_section_name):
        clues = [
            ('overload', cls.OVERLOAD),
            ('lunapark', cls.LUNAPARK),
        ]
        for clue, backend_type in clues:
            if clue in api_address:
                return backend_type
        else:
            for clue, backend_type in clues:
                if clue in cfg_section_name:
                    return backend_type
            else:
                raise KeyError(
                    'Can not identify backend: neither api address nor section name match any of the patterns:\n%s' %
                    '\n'.join(['*%s*' % ptrn[0] for ptrn in clues]))


class Job(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def push_test_data(self):
        """method for pushing test data to server"""

    @abstractmethod
    def edit_metainfo(self):
        """method for sending job metainfo to server"""

    @property
    @abstractmethod
    def number(self):
        """property with job number"""

    @abstractmethod
    def close(self):
        """method for closing job"""

    @abstractmethod
    def create(self):
        """method for creating job"""

    @abstractmethod
    def send_status(self):
        """method for sending status to server"""

    @abstractmethod
    def send_config(self):
        """method for sending test config to server"""

    @abstractmethod
    def push_monitoring_data(self):
        """method for pushing monitoring data"""

    @abstractmethod
    def push_events_data(self):
        """method for pushing event data"""

    @abstractmethod
    def lock_target(self):
        """method for locking target during..."""

    @abstractmethod
    def set_imbalance_and_dsc(self):
        """method for imbalance detection"""

    @abstractmethod
    def is_target_locked(self):
        """method for checking target lock"""


def chop(data_list, chunk_size):
    if sys.getsizeof(str(data_list)) <= chunk_size:
        return [data_list]
    elif len(data_list) == 1:
        logger.info("Too large piece of Telegraf data. Might experience upload problems.")
        return [data_list]
    else:
        mid = int(len(data_list) / 2)
        return chop(data_list[:mid], chunk_size) + chop(data_list[mid:], chunk_size)


class Plugin(AbstractPlugin, AggregateResultListener,
             MonitoringDataListener):
    RC_STOP_FROM_WEB = 8
    VERSION = '3.0'
    SECTION = 'uploader'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.data_queue = Queue()
        self.monitoring_queue = Queue()
        if self.core.error_log:
            self.events_queue = Queue()
            self.events_reader = EventsReader(self.core.error_log)
            self.events_processing = Drain(self.events_reader, self.events_queue)
            self.add_cleanup(self.stop_events_processing)
            self.events_processing.start()
            self.events = threading.Thread(target=self.__events_uploader)
            self.events.daemon = True

        self.retcode = -1
        self._target = None
        self.task_name = ''
        self.token_file = None
        self.version_tested = None
        self.send_status_period = 10

        self.status_sender = threading.Thread(target=self.__send_status)
        self.status_sender.daemon = True

        self.upload = threading.Thread(target=self.__data_uploader)
        self.upload.daemon = True

        self.monitoring = threading.Thread(target=self.__monitoring_uploader)
        self.monitoring.daemon = True

        self._is_telegraf = None
        self.backend_type = BackendTypes.identify_backend(self.cfg['api_address'], self.cfg_section_name)
        self._task = None
        self._api_token = ''
        self._lp_job = None
        self._lock_duration = None
        self._info = None
        self.locked_targets = []
        self.web_link = None
        self.finished = False

    def set_option(self, option, value):
        self.cfg.setdefault('meta', {})[option] = value
        self.core.publish(self.SECTION, 'meta.{}'.format(option), value)

    @staticmethod
    def get_key():
        return __file__

    @property
    def lock_duration(self):
        if self._lock_duration is None:
            info = self.get_generator_info()
            self._lock_duration = info.duration if info.duration else \
                expand_to_seconds(self.get_option("target_lock_duration"))
        return self._lock_duration

    def get_available_options(self):
        opts = [
            "api_address",
            "writer_endpoint",
            "task",
            "job_name",
            "job_dsc",
            "notify",
            "ver", "component",
            "operator",
            "jobno_file",
            "ignore_target_lock",
            "target_lock_duration",
            "lock_targets",
            "jobno",
            "upload_token",
            'connection_timeout',
            'network_attempts',
            'api_attempts',
            'maintenance_attempts',
            'network_timeout',
            'api_timeout',
            'maintenance_timeout',
            'strict_lock',
            'send_status_period',
            'log_data_requests',
            'log_monitoring_requests',
            'log_status_requests',
            'log_other_requests',
            'threads_timeout',
            'chunk_size'
        ]
        return opts

    def configure(self):
        self.core.publish(self.SECTION, 'component', self.get_option('component'))
        self.core.publish(self.SECTION, 'task', self.get_option('task'))
        self.core.publish(self.SECTION, 'job_name', self.get_option('job_name'))

    def check_task_is_open(self):
        if self.backend_type != BackendTypes.LUNAPARK:
            return
        TASK_TIP = 'The task should be connected to Lunapark.' \
                   'Open startrek task page, click "actions" -> "load testing".'

        logger.debug("Check if task %s is open", self.task)
        try:
            task_data = self.lp_job.get_task_data(self.task)[0]
            try:
                task_status = task_data['status']
                if task_status == 'Open':
                    logger.info("Task %s is ok", self.task)
                    self.task_name = str(task_data['name'])
                else:
                    logger.info("Task %s:", self.task)
                    logger.info(task_data)
                    raise RuntimeError("Task is not open")
            except KeyError:
                try:
                    error = task_data['error']
                    raise RuntimeError("Task %s error: %s\n%s", self.task, error, TASK_TIP)
                except KeyError:
                    raise RuntimeError('Unknown task data format:\n{}'.format(task_data))
        except requests.exceptions.HTTPError as ex:
            logger.error(
                "Failed to check task status for '%s': %s", self.task, ex)
            if ex.response.status_code == 404:
                raise RuntimeError("Task not found: %s\n%s" % (self.task, TASK_TIP))
            elif ex.response.status_code == 500 or ex.response.status_code == 400:
                raise RuntimeError(
                    "Unable to check task staus, id: %s, error code: %s" %
                    (self.task, ex.response.status_code))
            raise ex

    @staticmethod
    def search_task_from_cwd(cwd):
        issue = re.compile("^([A-Za-z]+-[0-9]+)(-.*)?")
        while cwd:
            logger.debug("Checking if dir is named like JIRA issue: %s", cwd)
            if issue.match(os.path.basename(cwd)):
                res = re.search(issue, os.path.basename(cwd))
                return res.group(1).upper()

            newdir = os.path.abspath(os.path.join(cwd, os.path.pardir))
            if newdir == cwd:
                break
            else:
                cwd = newdir

        raise RuntimeError("task=dir requested, but no JIRA issue name in cwd: %s" % os.getcwd())

    def prepare_test(self):
        info = self.get_generator_info()
        port = info.port
        instances = info.instances
        if info.ammo_file is not None:
            if info.ammo_file.startswith("http://") or info.ammo_file.startswith("https://"):
                ammo_path = info.ammo_file
            else:
                ammo_path = os.path.realpath(info.ammo_file)
        else:
            logger.warning('Failed to get info about ammo path')
            ammo_path = 'Undefined'
        loop_count = int(info.loop_count)

        try:
            lp_job = self.lp_job
            self.add_cleanup(self.unlock_targets)
            self.locked_targets = self.check_and_lock_targets(strict=self.get_option('strict_lock'),
                                                              ignore=self.get_option('ignore_target_lock'))
            if lp_job._number:
                self.make_symlink(lp_job._number)
                self.check_task_is_open()
            else:
                self.check_task_is_open()
                lp_job.create()
                self.make_symlink(lp_job.number)
            self.publish('job_no', lp_job.number)
        except (APIClient.JobNotCreated, APIClient.NotAvailable, APIClient.NetworkError):
            logger.exception('Failed to connect to Lunapark, disabling DataUploader')
            self.start_test = lambda *a, **kw: None
            self.post_process = lambda *a, **kw: None
            self.on_aggregated_data = lambda *a, **kw: None
            self.monitoring_data = lambda *a, **kw: None
            return

        cmdline = ' '.join(sys.argv)
        lp_job.edit_metainfo(
            instances=instances,
            ammo_path=ammo_path,
            loop_count=loop_count,
            regression_component=self.get_option("component"),
            cmdline=cmdline,
        )

        self.core.job.subscribe_plugin(self)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
            console = None

        if console:
            console.add_info_widget(JobInfoWidget(self))

        self.set_option('target_host', self.target)
        self.set_option('target_port', port)
        self.set_option('cmdline', cmdline)
        self.set_option('ammo_path', ammo_path)
        self.set_option('loop_count', loop_count)
        self.__save_conf()

    def start_test(self):
        self.add_cleanup(self.join_threads)
        self.status_sender.start()
        self.upload.start()
        self.monitoring.start()
        if self.core.error_log:
            self.events.start()

        self.web_link = urljoin(self.lp_job.api_client.base_url, str(self.lp_job.number))
        logger.info("Web link: %s", self.web_link)

        self.publish("jobno", self.lp_job.number)
        self.publish("web_link", self.web_link)

        jobno_file = self.get_option("jobno_file", '')
        if jobno_file:
            logger.debug("Saving jobno to: %s", jobno_file)
            with open(jobno_file, 'w') as fdes:
                fdes.write(str(self.lp_job.number))
            self.core.add_artifact_file(jobno_file)
        self.__save_conf()

    def is_test_finished(self):
        return self.retcode

    def end_test(self, retcode):
        if retcode != 0:
            self.lp_job.interrupted.set()
        self.__save_conf()
        self.unlock_targets()
        return retcode

    def close_job(self):
        self.lp_job.close(self.retcode)

    def join_threads(self):
        self.lp_job.interrupted.set()
        if self.monitoring.is_alive():
            self.monitoring.join()
        if self.upload.is_alive():
            self.upload.join()

    def stop_events_processing(self):
        self.events_queue.put(None)
        self.events_reader.close()
        self.events_processing.close()
        if self.events_processing.is_alive():
            self.events_processing.join()
        if self.events.is_alive():
            self.lp_job.interrupted.set()
            self.events.join()

    def post_process(self, rc):
        self.retcode = rc
        self.monitoring_queue.put(None)
        self.data_queue.put(None)
        if self.core.error_log:
            self.events_queue.put(None)
            self.events_reader.close()
            self.events_processing.close()
            self.events.join()
        logger.info("Waiting for sender threads to join.")
        if self.monitoring.is_alive():
            self.monitoring.join()
        if self.upload.is_alive():
            self.upload.join()
        self.finished = True
        logger.info("Web link: %s", self.web_link)
        autostop = None
        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
        except KeyError as ex:
            logger.debug(ex)

        if autostop and autostop.cause_criterion and autostop.imbalance_timestamp:
            self.lp_job.set_imbalance_and_dsc(
                autostop.imbalance_rps,
                autostop.cause_criterion.explain(),
                autostop.imbalance_timestamp
            )

        else:
            logger.debug("No autostop cause detected")
        self.__save_conf()
        return rc

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """
        if not self.lp_job.interrupted.is_set():
            self.data_queue.put((data, stats))

    def monitoring_data(self, data_list):
        if not self.lp_job.interrupted.is_set():
            if len(data_list) > 0:
                [self.monitoring_queue.put(chunk) for chunk in chop(data_list, self.get_option("chunk_size"))]

    def __send_status(self):
        logger.info('Status sender thread started')
        lp_job = self.lp_job
        while not lp_job.interrupted.is_set():
            try:
                self.lp_job.send_status(self.core.info.get_info_dict())
                time.sleep(self.get_option('send_status_period'))
            except (APIClient.NetworkError, APIClient.NotAvailable) as e:
                logger.warning('Failed to send status')
                logger.debug(e)
                break
            except APIClient.StoppedFromOnline:
                logger.info("Test stopped from Lunapark")
                self.retcode = self.RC_STOP_FROM_WEB
                break
            if self.finished:
                break
        logger.info("Closed Status sender thread")

    def __uploader(self, queue, sender_method, name='Uploader'):
        logger.info('{} thread started'.format(name))
        while not self.lp_job.interrupted.is_set():
            try:
                entry = queue.get(timeout=1)
                if entry is None:
                    logger.info("{} queue returned None".format(name))
                    break
                sender_method(entry)
            except Empty:
                continue
            except APIClient.StoppedFromOnline:
                logger.warning("Lunapark is rejecting {} data".format(name))
                break
            except (APIClient.NetworkError, APIClient.NotAvailable, APIClient.UnderMaintenance):
                logger.exception('Failed to push {} data'.format(name))
                self.lp_job.interrupted.set()
            except Exception:
                logger.exception("Unhandled exception occured. Skipping data chunk...")
        # purge queue
        while not queue.empty():
            if queue.get_nowait() is None:
                break
        logger.info("Closing {} thread".format(name))

    def __data_uploader(self):
        self.__uploader(self.data_queue,
                        lambda entry: self.lp_job.push_test_data(*entry),
                        'Data Uploader')

    def __monitoring_uploader(self):
        self.__uploader(self.monitoring_queue,
                        self.lp_job.push_monitoring_data,
                        'Monitoring Uploader')

    def __events_uploader(self):
        self.__uploader(self.events_queue,
                        self.lp_job.push_events_data,
                        'Events Uploader')

    # TODO: why we do it here? should be in core
    def __save_conf(self):
        for requisites, content in self.core.artifacts_to_send:
            self.lp_job.send_config(requisites, content)

    def parse_lock_targets(self):
        # prepare target lock list
        locks_list_cfg = self.get_option('lock_targets', 'auto')

        def no_target():
            logging.warn("Target lock set to 'auto', but no target info available")
            return {}

        locks_set = {self.target} or no_target() if locks_list_cfg == 'auto' else set(locks_list_cfg)
        targets_to_lock = [host for host in locks_set if host]
        return targets_to_lock

    def lock_targets(self, targets_to_lock, ignore, strict):
        locked_targets = [target for target in targets_to_lock
                          if self.lp_job.lock_target(target, self.lock_duration, ignore, strict)]
        return locked_targets

    def unlock_targets(self):
        logger.info("Unlocking targets: %s", self.locked_targets)
        for target in self.locked_targets:
            logger.info(target)
            self.lp_job.api_client.unlock_target(target)

    def check_and_lock_targets(self, strict, ignore):
        targets_list = self.parse_lock_targets()
        logger.info('Locking targets: %s', targets_list)
        locked_targets = self.lock_targets(targets_list, ignore=ignore, strict=strict)
        logger.info('Locked targets: %s', locked_targets)
        return locked_targets

    def make_symlink(self, name):
        PLUGIN_DIR = os.path.join(self.core.artifacts_base_dir, 'lunapark')
        if not os.path.exists(PLUGIN_DIR):
            os.makedirs(PLUGIN_DIR)
        try:
            os.symlink(
                os.path.relpath(
                    self.core.artifacts_dir,
                    PLUGIN_DIR),
                os.path.join(
                    PLUGIN_DIR,
                    str(name)))
        # this exception catch for filesystems w/o symlinks
        except OSError:
            logger.warning('Unable to create symlink for artifact: %s', name)

    def _get_user_agent(self):
        plugin_agent = 'Uploader/{}'.format(self.VERSION)
        return ' '.join((plugin_agent,
                         self.core.get_user_agent()))

    def __get_operator(self):
        try:
            return self.get_option(
                'operator') or pwd.getpwuid(
                os.geteuid())[0]
        except:  # noqa: E722
            logger.error(
                "Couldn't get username from the OS. Please, set the 'meta.operator' option explicitly in your config "
                "file.")
            raise

    def __get_api_client(self):
        logging.info('Using {} backend'.format(self.backend_type))
        self._api_token = None
        if self.backend_type == BackendTypes.LUNAPARK:
            client = APIClient
        elif self.backend_type == BackendTypes.OVERLOAD:
            client = OverloadClient
            self._api_token = self.read_token(self.get_option("token_file"))
        else:
            raise RuntimeError("Backend type doesn't match any of the expected")

        return client(base_url=self.get_option('api_address'),
                      writer_url=self.get_option('writer_endpoint'),
                      network_attempts=self.get_option('network_attempts'),
                      api_attempts=self.get_option('api_attempts'),
                      maintenance_attempts=self.get_option('maintenance_attempts'),
                      network_timeout=self.get_option('network_timeout'),
                      api_timeout=self.get_option('api_timeout'),
                      maintenance_timeout=self.get_option('maintenance_timeout'),
                      connection_timeout=self.get_option('connection_timeout'),
                      user_agent=self._get_user_agent(),
                      api_token=self.api_token,
                      core_interrupted=self.interrupted)

    @property
    def lp_job(self):
        """

        :rtype: LPJob
        """
        if self._lp_job is None:
            self._lp_job = self.__get_lp_job()
            if isinstance(self._lp_job, LPJob):
                self.core.publish(self.SECTION, 'job_no', self._lp_job.number)
                self.core.publish(self.SECTION, 'web_link', self._lp_job.web_link)
                self.core.publish(self.SECTION, 'job_name', self._lp_job.name)
                self.core.publish(self.SECTION, 'job_dsc', self._lp_job.description)
                self.core.publish(self.SECTION, 'person', self._lp_job.person)
                self.core.publish(self.SECTION, 'task', self._lp_job.task)
                self.core.publish(self.SECTION, 'version', self._lp_job.version)
                self.core.publish(self.SECTION, 'component', self.get_option('component'))
                self.core.publish(self.SECTION, 'meta', self.cfg.get('meta', {}))
        return self._lp_job

    def __get_lp_job(self):
        """

        :rtype: LPJob
        """
        api_client = self.__get_api_client()

        info = self.get_generator_info()
        port = info.port
        loadscheme = [] if isinstance(info.rps_schedule, (str, dict)) else info.rps_schedule

        lp_job = LPJob(client=api_client,
                       target_host=self.target,
                       target_port=port,
                       number=self.cfg.get('jobno'),
                       token=self.get_option('upload_token'),
                       person=self.__get_operator(),
                       task=self.task,
                       name=self.get_option('job_name', 'untitled'),
                       description=self.get_option('job_dsc'),
                       tank=self.core.job.tank,
                       notify_list=self.get_option("notify"),
                       load_scheme=loadscheme,
                       version=self.get_option('ver'),
                       log_data_requests=self.get_option('log_data_requests'),
                       log_monitoring_requests=self.get_option('log_monitoring_requests'),
                       log_status_requests=self.get_option('log_status_requests'),
                       log_other_requests=self.get_option('log_other_requests'),
                       add_cleanup=lambda: self.add_cleanup(self.close_job))
        lp_job.send_config(LPRequisites.CONFIGINITIAL, yaml.dump(self.core.configinitial))
        return lp_job

    @property
    def task(self):
        if self._task is None:
            task = self.get_option('task')
            if task == 'dir':
                task = self.search_task_from_cwd(os.getcwd())
            self._task = task
        return self._task

    @property
    def api_token(self):
        if self._api_token == '':
            if self.backend_type == BackendTypes.LUNAPARK:
                self._api_token = None
            elif self.backend_type == BackendTypes.OVERLOAD:
                self._api_token = self.read_token(self.get_option("token_file", ""))
            else:
                raise RuntimeError("Backend type doesn't match any of the expected")
        return self._api_token

    @staticmethod
    def read_token(filename):
        if filename:
            logger.debug("Trying to read token from %s", filename)
            try:
                with open(filename, 'r') as handle:
                    data = handle.read().strip()
                    logger.info(
                        "Read authentication token from %s, "
                        "token length is %d bytes", filename, len(str(data)))
            except IOError:
                logger.error("Failed to read Overload API token from %s", filename)
                logger.info(
                    "Get your Overload API token from https://overload.yandex.net and provide it via 'overload.token_file' parameter"
                )
                raise RuntimeError("API token error")
            return data
        else:
            logger.error("Overload API token filename is not defined")
            logger.info(
                "Get your Overload API token from https://overload.yandex.net and provide it via 'overload.token_file' parameter"
            )
            raise RuntimeError("API token error")

    def get_generator_info(self):
        return self.core.job.generator_plugin.get_info()

    @property
    def target(self):
        if self._target is None:
            self._target = self.get_generator_info().address
            logger.info("Detected target: %s", self._target)
        return self._target


class JobInfoWidget(AbstractInfoWidget):
    def __init__(self, sender):
        # type: (Plugin) -> object
        AbstractInfoWidget.__init__(self)
        self.owner = sender

    def get_index(self):
        return 1

    def render(self, screen):
        template = "Author: " + screen.markup.RED + "%s" + \
                   screen.markup.RESET + \
                   "%s\n   Job: %s %s\n  Task: %s %s\n   Web: %s%s"
        data = (self.owner.lp_job.person[:1], self.owner.lp_job.person[1:],
                self.owner.lp_job.number, self.owner.lp_job.name, self.owner.lp_job.task,
                # todo: task_name from api_client.get_task_data()
                self.owner.lp_job.task, self.owner.lp_job.api_client.base_url,
                self.owner.lp_job.number)

        return template % data


class LPJob(Job):
    def __init__(
        self,
        client,
        target_host,
        target_port,
        person,
        task,
        name,
        description,
        tank,
        log_data_requests=False,
        log_other_requests=False,
        log_status_requests=False,
        log_monitoring_requests=False,
        number=None,
        token=None,
        notify_list=None,
        version=None,
        detailed_time=None,
        load_scheme=None,
        add_cleanup=lambda: None
    ):
        """
        :param client: APIClient
        :param log_data_requests: bool
        :param log_other_request: bool
        :param log_status_requests: bool
        :param log_monitoring_requests: bool
        """
        assert bool(number) == bool(
            token), 'Job number and upload token should come together'
        self.log_other_requests = log_other_requests
        self.log_data_requests = log_data_requests
        self.log_status_requests = log_status_requests
        self.log_monitoring_requests = log_monitoring_requests
        self.name = name
        self.tank = tank
        self.target_host = target_host
        self.target_port = target_port
        self.person = person
        self.task = task
        self.interrupted = threading.Event()
        self._number = number
        self._token = token
        self.api_client = client
        self.notify_list = notify_list
        self.description = description
        self.version = version
        self.detailed_time = detailed_time
        self.load_scheme = load_scheme
        self.is_finished = False
        self.web_link = ''
        self.add_cleanup = add_cleanup
        if self._number:
            self.add_cleanup()

    def push_test_data(self, data, stats):
        if not self.interrupted.is_set():
            try:
                self.api_client.push_test_data(
                    self.number, self.token, data, stats, self.interrupted, trace=self.log_data_requests)
            except (APIClient.NotAvailable, APIClient.NetworkError, APIClient.UnderMaintenance):
                logger.warning('Failed to push test data')
                self.interrupted.set()

    def edit_metainfo(
        self,
        instances=0,
        ammo_path=None,
        loop_count=None,
        regression_component=None,
        cmdline=None,
        is_starred=False,
        tank_type=1
    ):
        try:
            self.api_client.edit_job_metainfo(jobno=self.number,
                                              job_name=self.name,
                                              job_dsc=self.description,
                                              instances=instances,
                                              ammo_path=ammo_path,
                                              loop_count=loop_count,
                                              version_tested=self.version,
                                              component=regression_component,
                                              cmdline=cmdline,
                                              is_starred=is_starred,
                                              tank_type=tank_type,
                                              trace=self.log_other_requests)
        except (APIClient.NotAvailable, APIClient.StoppedFromOnline, APIClient.NetworkError,
                APIClient.UnderMaintenance):
            logger.exception('Failed to edit job metainfo on Lunapark')

    @property
    def number(self):
        if not self._number:
            self.create()
        return self._number

    @property
    def token(self):
        if not self._token:
            self.create()
        return self._token

    def close(self, rc):
        if self._number:
            return self.api_client.close_job(self.number, rc, trace=self.log_other_requests)
        else:
            return True

    def create(self):
        self._number, self._token = self.api_client.new_job(task=self.task,
                                                            person=self.person,
                                                            tank=self.tank,
                                                            loadscheme=self.load_scheme,
                                                            target_host=self.target_host,
                                                            target_port=self.target_port,
                                                            detailed_time=self.detailed_time,
                                                            notify_list=self.notify_list,
                                                            trace=self.log_other_requests)
        self.add_cleanup()
        logger.info('Job created: {}'.format(self._number))
        self.web_link = urljoin(self.api_client.base_url, str(self._number))

    def send_status(self, status):
        if self._number and not self.interrupted.is_set():
            self.api_client.send_status(
                self.number,
                self.token,
                status,
                trace=self.log_status_requests)

    def get_task_data(self, task):
        return self.api_client.get_task_data(
            task, trace=self.log_other_requests)

    def send_config(self, lp_requisites, content):
        self.api_client.send_config(self.number, lp_requisites, content, trace=self.log_other_requests)

    def push_monitoring_data(self, data):
        if not self.interrupted.is_set():
            self.api_client.push_monitoring_data(
                self.number, self.token, data, self.interrupted, trace=self.log_monitoring_requests)

    def push_events_data(self, data):
        if not self.interrupted.is_set():
            self.api_client.push_events_data(self.number, self.person, data)

    def lock_target(self, lock_target, lock_target_duration, ignore, strict):
        lock_wait_timeout = 10
        maintenance_timeouts = iter([0]) if ignore else iter(lambda: lock_wait_timeout, 0)
        while True:
            try:
                self.api_client.lock_target(lock_target,
                                            lock_target_duration,
                                            trace=self.log_other_requests,
                                            maintenance_timeouts=maintenance_timeouts,
                                            maintenance_msg="Target is locked.\nManual unlock link: %s/%s" % (
                                                self.api_client.base_url,
                                                self.api_client.get_manual_unlock_link(lock_target)
                                            ))
                return True
            except (APIClient.NotAvailable, APIClient.StoppedFromOnline) as e:
                logger.info('Target is not locked due to %s', e)
                if ignore:
                    logger.info('ignore_target_locks = 1')
                    return False
                elif strict:
                    raise e
                else:
                    logger.info('strict_lock = 0')
                    return False
            except APIClient.UnderMaintenance:
                logger.info('Target is locked')
                if ignore:
                    logger.info('ignore_target_locks = 1')
                    return False
                logger.info("Manual unlock link: %s/%s",
                            self.api_client.base_url,
                            self.api_client.get_manual_unlock_link(lock_target))
                continue

    def set_imbalance_and_dsc(self, rps, comment, timestamp):
        return self.api_client.set_imbalance_and_dsc(self.number, rps, comment, timestamp)

    def is_target_locked(self, host, strict):
        while True:
            try:
                return self.api_client.is_target_locked(
                    host, trace=self.log_other_requests)
            except APIClient.UnderMaintenance:
                logger.info('Target is locked, retrying...')
                continue
            except (APIClient.StoppedFromOnline, APIClient.NotAvailable, APIClient.NetworkError):
                logger.info('Can\'t check whether target is locked\n')
                if strict:
                    logger.warning('Stopping test due to strict_lock')
                    raise
                else:
                    logger.warning('strict_lock is False, proceeding')
                    return {'status': 'ok'}


class EventsReader(FileScanner):
    """
    Parse lines and return stats
    """

    def __init__(self, *args, **kwargs):
        super(EventsReader, self).__init__(*args, **kwargs)

    def _read_data(self, lines):
        results = []
        for line in lines:
            # 2018-03-30 13:40:50,541\tCan't get monitoring config
            data = line.split("\t")
            if len(data) > 1:
                timestamp, message = data[0], data[1]
                dt = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S,%f')
                unix_ts = int(time.mktime(dt.timetuple()))
                results.append([unix_ts, message])
        return results
